"""Tests for message bus functionality."""

from datetime import datetime

from ya_agent_sdk.context import BusMessage, MessageBus

# =============================================================================
# BusMessage Tests
# =============================================================================


def test_bus_message_creation() -> None:
    """Test BusMessage creation with required fields."""
    msg = BusMessage(id="msg-001", content="Hello", source="user")
    assert msg.id == "msg-001"
    assert msg.content == "Hello"
    assert msg.source == "user"
    assert msg.target is None
    assert isinstance(msg.timestamp, datetime)


def test_bus_message_auto_generates_id() -> None:
    """Test BusMessage auto-generates UUID id if not provided."""
    msg = BusMessage(content="Hello", source="user")
    assert len(msg.id) == 32  # UUID hex format


def test_bus_message_with_target() -> None:
    """Test BusMessage creation with target."""
    msg = BusMessage(id="msg-001", content="Hello", source="user", target="main")
    assert msg.target == "main"


def test_bus_message_render_without_template() -> None:
    """Test message rendering without template."""
    msg = BusMessage(id="msg-001", content="Hello", source="user")
    assert msg.render() == "Hello"


def test_bus_message_render_with_template() -> None:
    """Test message rendering with Jinja2 template."""
    msg = BusMessage(id="msg-001", content="Stop", source="user", template="[URGENT] {{ content }}")
    assert msg.render() == "[URGENT] Stop"


def test_bus_message_empty_id_rejected() -> None:
    """Test that empty id is rejected."""
    import pytest

    with pytest.raises(ValueError, match="id must be non-empty"):
        BusMessage(id="", content="Hello", source="user")

    with pytest.raises(ValueError, match="id must be non-empty"):
        BusMessage(id="   ", content="Hello", source="user")


# =============================================================================
# MessageBus Subscribe/Unsubscribe Tests
# =============================================================================


def test_message_bus_subscribe() -> None:
    """Test subscribing to the bus."""
    bus = MessageBus()
    assert bus.subscriber_count == 0

    bus.subscribe("main")
    assert bus.subscriber_count == 1

    # Idempotent
    bus.subscribe("main")
    assert bus.subscriber_count == 1


def test_message_bus_unsubscribe() -> None:
    """Test unsubscribing from the bus."""
    bus = MessageBus()
    bus.subscribe("main")
    bus.subscribe("subagent")
    assert bus.subscriber_count == 2

    bus.unsubscribe("subagent")
    assert bus.subscriber_count == 1

    # Unsubscribe non-existent is no-op
    bus.unsubscribe("nonexistent")
    assert bus.subscriber_count == 1


def test_message_bus_auto_subscribe_on_consume() -> None:
    """Test that consume auto-subscribes if not already subscribed."""
    bus = MessageBus()
    bus.send(BusMessage(content="Hello", source="user", target="main"))

    # Not subscribed yet
    assert bus.subscriber_count == 0

    # Consume auto-subscribes
    messages = bus.consume("main")
    assert bus.subscriber_count == 1
    # But since we subscribed at latest, we won't see the old message
    assert len(messages) == 0


# =============================================================================
# MessageBus Send/Consume Tests
# =============================================================================


def test_message_bus_send() -> None:
    """Test sending messages to the bus."""
    bus = MessageBus()
    msg = bus.send(BusMessage(content="Hello", source="user"))

    assert len(msg.id) == 32  # UUID hex
    assert msg.content == "Hello"
    assert msg.source == "user"
    assert len(bus) == 1


def test_message_bus_send_with_explicit_id() -> None:
    """Test sending messages with explicit id."""
    bus = MessageBus()
    msg = bus.send(BusMessage(id="custom-id-123", content="Hello", source="user"))

    assert msg.id == "custom-id-123"
    assert msg.content == "Hello"


def test_message_bus_send_generates_unique_ids() -> None:
    """Test that auto-generated message IDs are unique."""
    bus = MessageBus()
    msg1 = bus.send(BusMessage(content="First", source="user"))
    msg2 = bus.send(BusMessage(content="Second", source="user"))
    msg3 = bus.send(BusMessage(content="Third", source="user"))

    assert msg1.id != msg2.id != msg3.id


def test_message_bus_send_idempotent() -> None:
    """Test that send is idempotent with same message id."""
    bus = MessageBus()
    msg1 = bus.send(BusMessage(id="same-id", content="Hello", source="user"))
    msg2 = bus.send(BusMessage(id="same-id", content="Different", source="other"))

    # Should return the same message, not create duplicate
    assert msg1.id == msg2.id
    assert msg1.content == "Hello"  # Original content preserved
    assert len(bus) == 1


def test_message_bus_consume_targeted() -> None:
    """Test consuming messages targeted at specific agent."""
    bus = MessageBus()
    bus.subscribe("main")
    bus.subscribe("other")

    bus.send(BusMessage(content="For main", source="user", target="main"))
    bus.send(BusMessage(content="For other", source="user", target="other"))

    messages = bus.consume("main")
    assert len(messages) == 1
    assert messages[0].content == "For main"

    messages = bus.consume("other")
    assert len(messages) == 1
    assert messages[0].content == "For other"


def test_message_bus_consume_broadcast() -> None:
    """Test consuming broadcast messages (target=None)."""
    bus = MessageBus()
    bus.subscribe("main")
    bus.subscribe("other")

    bus.send(BusMessage(content="Broadcast", source="user"))  # No target = broadcast

    # Both subscribers receive the broadcast
    main_messages = bus.consume("main")
    assert len(main_messages) == 1
    assert main_messages[0].content == "Broadcast"

    other_messages = bus.consume("other")
    assert len(other_messages) == 1
    assert other_messages[0].content == "Broadcast"


def test_message_bus_consume_fifo_order() -> None:
    """Test that messages are consumed in FIFO order."""
    bus = MessageBus()
    bus.subscribe("main")

    bus.send(BusMessage(content="First", source="user", target="main"))
    bus.send(BusMessage(content="Second", source="user", target="main"))
    bus.send(BusMessage(content="Third", source="user", target="main"))

    messages = bus.consume("main")
    assert [m.content for m in messages] == ["First", "Second", "Third"]


def test_message_bus_consume_advances_cursor() -> None:
    """Test that cursor advances after consume."""
    bus = MessageBus()
    bus.subscribe("main")

    bus.send(BusMessage(content="First", source="user", target="main"))
    messages = bus.consume("main")
    assert len(messages) == 1

    # Second consume returns empty (already read)
    messages = bus.consume("main")
    assert len(messages) == 0

    # New message is visible
    bus.send(BusMessage(content="Second", source="user", target="main"))
    messages = bus.consume("main")
    assert len(messages) == 1
    assert messages[0].content == "Second"


def test_message_bus_has_pending() -> None:
    """Test has_pending check."""
    bus = MessageBus()
    bus.subscribe("main")

    assert not bus.has_pending("main")

    bus.send(BusMessage(content="Hello", source="user", target="main"))
    assert bus.has_pending("main")

    bus.consume("main")
    assert not bus.has_pending("main")


def test_message_bus_has_pending_broadcast() -> None:
    """Test has_pending with broadcast messages."""
    bus = MessageBus()
    bus.subscribe("main")
    bus.subscribe("other")

    bus.send(BusMessage(content="Broadcast", source="user"))  # No target

    assert bus.has_pending("main")
    assert bus.has_pending("other")


def test_message_bus_has_pending_unsubscribed() -> None:
    """Test has_pending for unsubscribed agent returns False."""
    bus = MessageBus()
    bus.send(BusMessage(content="Hello", source="user", target="main"))

    # Not subscribed, so has_pending is False
    assert not bus.has_pending("main")


def test_message_bus_peek() -> None:
    """Test peeking at messages without advancing cursor."""
    bus = MessageBus()
    bus.subscribe("main")
    bus.send(BusMessage(content="Hello", source="user", target="main"))

    messages = bus.peek("main")
    assert len(messages) == 1

    # Peek again still returns same messages
    messages = bus.peek("main")
    assert len(messages) == 1

    # Consume advances cursor
    bus.consume("main")
    messages = bus.peek("main")
    assert len(messages) == 0


def test_message_bus_peek_unsubscribed() -> None:
    """Test peek returns empty for unsubscribed agent."""
    bus = MessageBus()
    bus.send(BusMessage(content="Hello", source="user", target="main"))

    # Not subscribed, should return empty
    assert bus.peek("main") == []


def test_message_bus_clear() -> None:
    """Test clearing all messages and cursors."""
    bus = MessageBus()
    bus.subscribe("main")
    bus.send(BusMessage(content="One", source="user"))
    bus.send(BusMessage(content="Two", source="user"))

    assert len(bus) == 2
    assert bus.subscriber_count == 1

    bus.clear()
    assert len(bus) == 0
    assert bus.subscriber_count == 0


def test_message_bus_bool() -> None:
    """Test boolean evaluation of bus."""
    bus = MessageBus()
    assert not bus

    bus.send(BusMessage(content="Hello", source="user"))
    assert bus


# =============================================================================
# MessageBus Maxlen Tests
# =============================================================================


def test_message_bus_maxlen() -> None:
    """Test that old messages are trimmed when maxlen exceeded."""
    bus = MessageBus(maxlen=3)
    bus.subscribe("main")

    bus.send(BusMessage(content="One", source="user", target="main"))
    bus.send(BusMessage(content="Two", source="user", target="main"))
    bus.send(BusMessage(content="Three", source="user", target="main"))
    assert len(bus) == 3

    bus.send(BusMessage(content="Four", source="user", target="main"))
    assert len(bus) == 3

    messages = bus.consume("main")
    # "One" was trimmed
    assert [m.content for m in messages] == ["Two", "Three", "Four"]


def test_message_bus_cursor_correction_on_trim() -> None:
    """Test that cursor is corrected when it points to trimmed messages."""
    bus = MessageBus(maxlen=3)
    bus.subscribe("main")

    # Send and don't consume
    bus.send(BusMessage(content="One", source="user", target="main"))
    bus.send(BusMessage(content="Two", source="user", target="main"))
    bus.send(BusMessage(content="Three", source="user", target="main"))

    # Add more messages, trimming old ones
    bus.send(BusMessage(content="Four", source="user", target="main"))
    bus.send(BusMessage(content="Five", source="user", target="main"))

    # Cursor still at 0, but messages 1-2 are trimmed
    # Should get messages from earliest available
    messages = bus.consume("main")
    assert [m.content for m in messages] == ["Three", "Four", "Five"]


def test_message_bus_idempotency_does_not_survive_trim() -> None:
    """Test that idempotency does NOT survive trim.

    After a message is trimmed, its ID is removed from tracking.
    Resending creates a new message. This is expected behavior to bound memory.
    """
    bus = MessageBus(maxlen=2)

    # Send message that will be trimmed
    bus.send(BusMessage(id="id-001", content="First", source="user"))
    bus.send(BusMessage(id="id-002", content="Second", source="user"))
    bus.send(BusMessage(id="id-003", content="Third", source="user"))

    # id-001 is trimmed AND removed from _message_ids
    assert len(bus) == 2

    # Resending id-001 is NOT idempotent - treated as new message
    msg_retry = bus.send(BusMessage(id="id-001", content="First retry", source="user"))
    assert msg_retry.id == "id-001"
    assert msg_retry.content == "First retry"  # New content accepted
    assert len(bus) == 2  # Still 2 because id-002 was trimmed


# =============================================================================
# MessageBus Multi-Agent Tests
# =============================================================================


def test_message_bus_subagent_lifecycle() -> None:
    """Test typical subagent subscribe/unsubscribe lifecycle."""
    bus = MessageBus()
    bus.subscribe("main")

    # Send message to main
    bus.send(BusMessage(content="For main", source="user", target="main"))

    # Subagent starts, subscribes
    bus.subscribe("subagent-123")

    # Broadcast goes to both
    bus.send(BusMessage(content="Broadcast", source="user"))

    # Main gets both messages
    main_msgs = bus.consume("main")
    assert len(main_msgs) == 2

    # Subagent only gets broadcast (subscribed after first message)
    sub_msgs = bus.consume("subagent-123")
    assert len(sub_msgs) == 1
    assert sub_msgs[0].content == "Broadcast"

    # Subagent exits, unsubscribes
    bus.unsubscribe("subagent-123")
    assert bus.subscriber_count == 1


def test_message_bus_targeted_inter_agent() -> None:
    """Test targeted messages between agents."""
    bus = MessageBus()
    bus.subscribe("main")
    bus.subscribe("subagent")

    # Main sends to subagent
    bus.send(BusMessage(content="Task for you", source="main", target="subagent"))
    # Subagent sends to main
    bus.send(BusMessage(content="Result", source="subagent", target="main"))

    main_msgs = bus.consume("main")
    assert len(main_msgs) == 1
    assert main_msgs[0].content == "Result"

    sub_msgs = bus.consume("subagent")
    assert len(sub_msgs) == 1
    assert sub_msgs[0].content == "Task for you"


def test_message_bus_new_subscriber_only_sees_new_messages() -> None:
    """Test that new subscribers only see messages sent after subscribing."""
    bus = MessageBus()
    bus.subscribe("main")

    # Send before subagent subscribes
    bus.send(BusMessage(content="Old message", source="user", target="main"))

    # Subagent subscribes (should not see old broadcast)
    bus.subscribe("subagent")

    # New broadcast
    bus.send(BusMessage(content="New broadcast", source="user"))

    # Subagent only sees new message
    sub_msgs = bus.consume("subagent")
    assert len(sub_msgs) == 1
    assert sub_msgs[0].content == "New broadcast"


# =============================================================================
# Multimodal Content Tests
# =============================================================================


def test_bus_message_multimodal_content() -> None:
    """Test BusMessage with multimodal content (Sequence[UserContent])."""
    from pydantic_ai.messages import ImageUrl

    msg = BusMessage(
        id="mm-001",
        content=["Check this image:", ImageUrl(url="https://example.com/img.png")],
        source="user",
    )
    assert msg.id == "mm-001"
    assert isinstance(msg.content, list)
    assert len(msg.content) == 2
    assert msg.content[0] == "Check this image:"
    assert isinstance(msg.content[1], ImageUrl)


def test_bus_message_multimodal_render_ignores_template() -> None:
    """Test that render() ignores template for multimodal content."""
    from pydantic_ai.messages import ImageUrl

    msg = BusMessage(
        content=["Look at this:", ImageUrl(url="https://example.com/img.png")],
        source="user",
        template="[URGENT] {{ content }}",  # Should be ignored
    )
    rendered = msg.render()
    # Multimodal content returned as-is, template ignored
    assert isinstance(rendered, list)
    assert len(rendered) == 2
    assert rendered[0] == "Look at this:"


def test_bus_message_multimodal_render_without_template() -> None:
    """Test that render() returns multimodal content as-is without template."""
    from pydantic_ai.messages import AudioUrl

    content = ["Listen:", AudioUrl(url="https://example.com/audio.mp3")]
    msg = BusMessage(content=content, source="user")
    rendered = msg.render()
    assert rendered == content  # Same value, no transformation


def test_bus_message_content_text_str() -> None:
    """Test content_text() returns the string for str content."""
    msg = BusMessage(content="Hello world", source="user")
    assert msg.content_text() == "Hello world"


def test_bus_message_content_text_multimodal() -> None:
    """Test content_text() extracts text parts and shows placeholders for media."""
    from pydantic_ai.messages import ImageUrl

    msg = BusMessage(
        content=["Check this:", ImageUrl(url="https://example.com/img.png"), "What do you think?"],
        source="user",
    )
    text = msg.content_text()
    assert "Check this:" in text
    assert "[image-url]" in text
    assert "What do you think?" in text


def test_bus_message_content_text_empty_multimodal() -> None:
    """Test content_text() returns empty string for empty multimodal content."""
    msg = BusMessage(content=[], source="user")
    assert msg.content_text() == ""


def test_content_as_text_str() -> None:
    """Test content_as_text utility with str input."""
    from ya_agent_sdk.context.bus import content_as_text

    assert content_as_text("hello") == "hello"


def test_content_as_text_multimodal() -> None:
    """Test content_as_text utility with multimodal input."""
    from pydantic_ai.messages import DocumentUrl, ImageUrl

    from ya_agent_sdk.context.bus import content_as_text

    result = content_as_text(["text part", ImageUrl(url="https://a.com/i.png"), DocumentUrl(url="https://a.com/d.pdf")])
    assert result == "text part [image-url] [document-url]"


def test_bus_message_multimodal_in_bus() -> None:
    """Test multimodal messages work with MessageBus send/consume."""
    from pydantic_ai.messages import ImageUrl

    bus = MessageBus()
    bus.subscribe("main")

    content = ["Check this:", ImageUrl(url="https://example.com/img.png")]
    bus.send(BusMessage(content=content, source="user", target="main"))

    messages = bus.consume("main")
    assert len(messages) == 1
    assert isinstance(messages[0].content, list)
    assert len(messages[0].content) == 2
    assert messages[0].content[0] == "Check this:"
    assert isinstance(messages[0].content[1], ImageUrl)



"""Test query tools."""

from __future__ import annotations

from ya_agent_sdk.toolsets.browser_use._tools import build_tool
from ya_agent_sdk.toolsets.browser_use.tools import find_elements, navigate_to_url
from ya_agent_sdk.toolsets.browser_use.tools.query import get_element_attributes, get_element_text
from ya_agent_sdk.toolsets.browser_use.toolset import BrowserUseToolset


async def test_find_elements(cdp_url, test_server):
    """Test finding elements on page."""
    async with BrowserUseToolset(cdp_url) as toolset:
        session = toolset._browser_session

        # Navigate first using tool
        nav_tool = build_tool(session, navigate_to_url)
        await nav_tool.function_schema.call({"url": f"{test_server}/test_fixtures/basic.html"}, None)

        # Find h1 elements
        tool = build_tool(session, find_elements)
        result = await tool.function_schema.call({"selector": "h1", "limit": 10}, None)

        assert result["status"] == "success"
        assert result["count"] >= 0
        assert "elements" in result

        if result["count"] > 0:
            element = result["elements"][0]
            assert "tag_name" in element
            assert element["tag_name"] == "h1"
            assert "text" in element


async def test_find_elements_not_found(cdp_url, test_server):
    """Test finding elements with selector that doesn't match."""
    async with BrowserUseToolset(cdp_url) as toolset:
        session = toolset._browser_session

        nav_tool = build_tool(session, navigate_to_url)
        await nav_tool.function_schema.call({"url": f"{test_server}/test_fixtures/basic.html"}, None)

        # Find non-existent elements
        tool = build_tool(session, find_elements)
        result = await tool.function_schema.call({"selector": ".non-existent-class-xyz", "limit": 10}, None)

        assert result["status"] == "success"
        assert result["count"] == 0
        assert len(result["elements"]) == 0


async def test_find_elements_with_limit(cdp_url, test_server):
    """Test finding elements with result limit."""
    async with BrowserUseToolset(cdp_url) as toolset:
        session = toolset._browser_session

        nav_tool = build_tool(session, navigate_to_url)
        await nav_tool.function_schema.call({"url": f"{test_server}/test_fixtures/basic.html"}, None)

        # Find all elements with small limit
        tool = build_tool(session, find_elements)
        result = await tool.function_schema.call({"selector": "*", "limit": 5}, None)

        assert result["status"] == "success"
        assert len(result["elements"]) <= 5


async def test_get_element_text(cdp_url, test_server):
    """Test getting element text content."""
    async with BrowserUseToolset(cdp_url) as toolset:
        session = toolset._browser_session

        nav_tool = build_tool(session, navigate_to_url)
        await nav_tool.function_schema.call({"url": f"{test_server}/test_fixtures/basic.html"}, None)

        # Get text from h1 element
        text_tool = build_tool(session, get_element_text)
        result = await text_tool.function_schema.call({"selector": "h1"}, None)

        assert isinstance(result, str)
        # May be empty if h1 doesn't exist, but should not raise error


async def test_get_element_attributes(cdp_url, test_server):
    """Test getting element attributes."""
    async with BrowserUseToolset(cdp_url) as toolset:
        session = toolset._browser_session

        nav_tool = build_tool(session, navigate_to_url)
        await nav_tool.function_schema.call({"url": f"{test_server}/test_fixtures/basic.html"}, None)

        # Get attributes from link element
        attr_tool = build_tool(session, get_element_attributes)
        result = await attr_tool.function_schema.call({"selector": "a", "attributes": None}, None)

        assert isinstance(result, dict)


async def test_get_element_attributes_specific(cdp_url, test_server):
    """Test getting specific attributes from element."""
    async with BrowserUseToolset(cdp_url) as toolset:
        session = toolset._browser_session

        nav_tool = build_tool(session, navigate_to_url)
        await nav_tool.function_schema.call({"url": f"{test_server}/test_fixtures/basic.html"}, None)

        # Get specific attributes
        attr_tool = build_tool(session, get_element_attributes)
        result = await attr_tool.function_schema.call({"selector": "a", "attributes": ["href"]}, None)

        assert isinstance(result, dict)



"""Shared HTTP client utilities with SSRF protection."""

from __future__ import annotations

import ipaddress
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from functools import lru_cache
from typing import TYPE_CHECKING, Any
from urllib.parse import urljoin, urlparse

import httpx

if TYPE_CHECKING:
    pass

# Blocked IP ranges for SSRF protection
_BLOCKED_RANGES: tuple[ipaddress.IPv4Network | ipaddress.IPv6Network, ...] = (
    ipaddress.IPv4Network("10.0.0.0/8"),
    ipaddress.IPv4Network("172.16.0.0/12"),
    ipaddress.IPv4Network("192.168.0.0/16"),
    ipaddress.IPv4Network("127.0.0.0/8"),
    ipaddress.IPv4Network("169.254.0.0/16"),
    ipaddress.IPv6Network("::1/128"),
    ipaddress.IPv6Network("fc00::/7"),
    ipaddress.IPv6Network("fe80::/10"),
)


class ForbiddenUrlError(Exception):
    """Raised when URL is forbidden due to security restrictions."""


def verify_url(url: str) -> None:
    """Verify URL is safe to access (not pointing to internal resources).

    Args:
        url: The URL to verify.

    Raises:
        ForbiddenUrlError: If URL points to internal/private resources.
    """
    parsed = urlparse(url)

    if parsed.scheme not in ("http", "https"):
        raise ForbiddenUrlError(f"Unsupported scheme: {parsed.scheme}")

    hostname = parsed.hostname
    if not hostname:
        raise ForbiddenUrlError("Missing hostname")

    # Check for localhost variations
    if hostname in ("localhost", "127.0.0.1", "::1", "0.0.0.0"):  # noqa: S104
        raise ForbiddenUrlError(f"Access to localhost is forbidden: {hostname}")

    # Try to parse as IP address and check against blocked ranges
    try:
        ip = ipaddress.ip_address(hostname)
        for blocked in _BLOCKED_RANGES:
            if ip in blocked:
                raise ForbiddenUrlError(f"Access to private IP range is forbidden: {hostname}")
    except ValueError:
        # Not an IP address, hostname is allowed
        pass


@lru_cache(maxsize=1)
def _get_http_client() -> httpx.AsyncClient:
    """Internal cached client factory."""
    # Connection pool limits: max 100 connections, 20 per host
    limits = httpx.Limits(max_connections=100, max_keepalive_connections=20)
    # Timeout: 5 minutes total, 30s connect
    timeout = httpx.Timeout(300.0, connect=30.0)

    return httpx.AsyncClient(
        timeout=timeout,
        limits=limits,
        follow_redirects=False,  # Manual redirect handling for SSRF protection
        headers={
            "User-Agent": "Mozilla/5.0 (compatible; ya-agent-sdk/1.0)",
        },
    )


def get_http_client() -> httpx.AsyncClient:
    """Get shared async HTTP client instance.

    Handles closed client scenario by invalidating cache and creating new instance.

    Returns:
        Shared httpx.AsyncClient instance.
    """
    client = _get_http_client()
    if client.is_closed:
        # Invalidate cache and create new instance
        _get_http_client.cache_clear()
        client = _get_http_client()
    return client


async def safe_request(
    url: str,
    method: str = "GET",
    max_redirects: int = 10,
    skip_verification: bool = False,
    **kwargs,
) -> httpx.Response:
    """Make HTTP request with SSRF protection and safe redirect following.

    Args:
        url: Target URL.
        method: HTTP method.
        max_redirects: Maximum redirects to follow.
        skip_verification: Skip SSRF URL verification (for internal network environments).
        **kwargs: Additional arguments passed to httpx.

    Returns:
        HTTP response.

    Raises:
        ForbiddenUrlError: If URL or redirect target is forbidden.
        httpx.HTTPError: For HTTP errors.
    """
    if not skip_verification:
        verify_url(url)
    client = get_http_client()
    current_url = url

    for _ in range(max_redirects):
        response = await client.request(method, current_url, **kwargs)

        if response.status_code in (301, 302, 303, 307, 308):
            location = response.headers.get("Location")
            if not location:
                raise httpx.HTTPStatusError(
                    "Redirect missing Location header",
                    request=response.request,
                    response=response,
                )

            # Handle relative redirects
            if not location.startswith(("http://", "https://")):
                location = urljoin(current_url, location)

            if not skip_verification:
                verify_url(location)
            current_url = location
            # For 303, always use GET
            if response.status_code == 303:
                method = "GET"
        else:
            return response

    raise httpx.TooManyRedirects(f"Exceeded {max_redirects} redirects")


@asynccontextmanager
async def safe_stream_request(
    url: str,
    method: str = "GET",
    max_redirects: int = 10,
    skip_verification: bool = False,
    **kwargs: Any,
) -> AsyncIterator[httpx.Response]:
    """Make streamed HTTP request with SSRF protection and safe redirects."""
    if not skip_verification:
        verify_url(url)
    client = get_http_client()
    current_url = url
    request_method = method
    response: httpx.Response | None = None

    try:
        for _ in range(max_redirects):
            request = client.build_request(request_method, current_url, **kwargs)
            response = await client.send(request, stream=True)

            if response.status_code in (301, 302, 303, 307, 308):
                location = response.headers.get("Location")
                if not location:
                    raise httpx.HTTPStatusError(
                        "Redirect missing Location header",
                        request=response.request,
                        response=response,
                    )

                if not location.startswith(("http://", "https://")):
                    location = urljoin(current_url, location)

                if not skip_verification:
                    verify_url(location)

                current_url = location
                if response.status_code == 303:
                    request_method = "GET"

                await response.aclose()
                response = None
                continue

            yield response
            return

        raise httpx.TooManyRedirects(f"Exceeded {max_redirects} redirects")
    finally:
        if response is not None:
            await response.aclose()


async def check_url_accessible(url: str, timeout: float = 5.0, skip_verification: bool = False) -> bool:
    """Check if URL is accessible via HEAD request.

    Args:
        url: URL to check.
        timeout: Request timeout in seconds.
        skip_verification: Skip SSRF URL verification (for internal network environments).

    Returns:
        True if accessible, False otherwise.
    """
    try:
        response = await safe_request(url, method="HEAD", timeout=timeout, skip_verification=skip_verification)
        # Some servers don't support HEAD, try GET
        if response.status_code == 405:
            response = await safe_request(url, method="GET", timeout=timeout, skip_verification=skip_verification)
        response.raise_for_status()
        return True
    except (httpx.HTTPError, ForbiddenUrlError):
        return False



"""Web search tools: search, search_stock_image, search_image."""

from __future__ import annotations

import asyncio
from functools import cache
from pathlib import Path
from typing import TYPE_CHECKING, Annotated, Any

from pydantic import Field
from pydantic_ai import RunContext

from ya_agent_sdk.context import AgentContext
from ya_agent_sdk.toolsets.core.base import BaseTool
from ya_agent_sdk.toolsets.core.web._http_client import check_url_accessible, get_http_client

if TYPE_CHECKING:
    pass

URL_CHECK_TIMEOUT = 5.0
URL_VALIDATION_CONCURRENCY = 5
_PROMPTS_DIR = Path(__file__).parent / "prompts"


@cache
def _load_search_instruction() -> str:
    return (_PROMPTS_DIR / "search.md").read_text()


@cache
def _load_search_stock_image_instruction() -> str:
    return (_PROMPTS_DIR / "search_stock_image.md").read_text()


@cache
def _load_search_image_instruction() -> str:
    return (_PROMPTS_DIR / "search_image.md").read_text()


@cache
def _is_tavily_available() -> bool:
    """Check if tavily package is installed."""
    try:
        import tavily  # noqa: F401

        return True
    except ImportError:
        return False


class SearchTool(BaseTool):
    """Web search tool using Google, Brave, or Tavily."""

    name = "search"
    description = "Search the web for information using search APIs."

    def is_available(self, ctx: RunContext[AgentContext]) -> bool:
        """Available if Google, Brave, or Tavily API keys are configured."""
        cfg = ctx.deps.tool_config
        has_google = bool(cfg.google_search_api_key and cfg.google_search_cx)
        has_brave = bool(cfg.brave_search_api_key)
        has_tavily = bool(cfg.tavily_api_key) and _is_tavily_available()
        return has_google or has_brave or has_tavily

    async def get_instruction(self, ctx: RunContext[AgentContext]) -> str | None:
        if not self.is_available(ctx):
            return None
        return _load_search_instruction()

    async def call(
        self,
        ctx: RunContext[AgentContext],
        query: Annotated[str, Field(description="The search query")],
        num: Annotated[int, Field(description="Number of results to return (1-10)", default=10)] = 10,
    ) -> list[dict[str, Any]] | dict[str, Any]:
        """Execute web search."""
        cfg = ctx.deps.tool_config
        has_brave = bool(cfg.brave_search_api_key)
        has_tavily = bool(cfg.tavily_api_key) and _is_tavily_available()

        # Priority: Google > Brave > Tavily
        if cfg.google_search_api_key and cfg.google_search_cx:
            result = await self._search_google(query, num, cfg.google_search_api_key, cfg.google_search_cx)
            if isinstance(result, dict) and result.get("success") is False:
                # Fallback to Brave or Tavily
                if has_brave:
                    return await self._search_brave(query, num, cfg.brave_search_api_key)  # type: ignore[arg-type]
                if has_tavily:
                    return await self._search_tavily(query, cfg.tavily_api_key)  # type: ignore[arg-type]
            return result
        elif has_brave:
            return await self._search_brave(query, num, cfg.brave_search_api_key)  # type: ignore[arg-type]
        elif has_tavily:
            return await self._search_tavily(query, cfg.tavily_api_key)  # type: ignore[arg-type]
        else:
            return {"success": False, "error": "No search API available"}

    async def _search_google(
        self, query: str, num: int, api_key: str, cx: str
    ) -> list[dict[str, Any]] | dict[str, Any]:
        """Search using Google Custom Search API."""
        if not 1 <= num <= 10:
            return {"success": False, "error": "num must be between 1 and 10"}

        client = get_http_client()
        params = {
            "q": query,
            "num": num,
            "key": api_key,
            "cx": cx,
        }

        response = await client.get(
            "https://www.googleapis.com/customsearch/v1",
            params=params,
            timeout=60,
        )

        if response.status_code != 200:
            return {"success": False, "error": f"HTTP {response.status_code}: {response.text}"}

        return response.json()

    async def _search_brave(self, query: str, num: int, api_key: str) -> list[dict[str, Any]] | dict[str, Any]:
        """Search using Brave Search API."""
        client = get_http_client()
        # Brave API max count is 20
        params = {"q": query, "count": min(num, 20)}
        headers = {"X-Subscription-Token": api_key}

        response = await client.get(
            "https://api.search.brave.com/res/v1/web/search",
            params=params,
            headers=headers,
            timeout=60,
        )

        if response.status_code != 200:
            return {"success": False, "error": f"HTTP {response.status_code}: {response.text}"}

        data = response.json()
        # Extract web results from Brave response
        web_results = data.get("web", {}).get("results", [])
        if not web_results:
            return {"success": False, "error": "No search results found."}

        # Normalize to simple format
        return [
            {
                "title": r.get("title", ""),
                "url": r.get("url", ""),
                "description": r.get("description", ""),
            }
            for r in web_results
        ]

    async def _search_tavily(self, query: str, api_key: str) -> list[dict[str, Any]] | dict[str, Any]:
        """Search using Tavily API."""
        from tavily import AsyncTavilyClient

        client = AsyncTavilyClient(api_key)
        results = await client.search(query, search_depth="advanced")  # type: ignore[arg-type]

        if not results.get("results"):
            return {"success": False, "error": "No search results found."}

        return results["results"]


class SearchStockImageTool(BaseTool):
    """Stock image search using Pixabay."""

    name = "search_stock_image"
    description = "Search royalty-free stock images from Pixabay for design work."

    def is_available(self, ctx: RunContext[AgentContext]) -> bool:
        """Available if Pixabay API key is configured."""
        return bool(ctx.deps.tool_config.pixabay_api_key)

    async def get_instruction(self, ctx: RunContext[AgentContext]) -> str | None:
        return _load_search_stock_image_instruction()

    async def call(
        self,
        ctx: RunContext[AgentContext],
        query: Annotated[
            str,
            Field(description="Search term (max 100 chars). E.g., 'business team', 'nature landscape'"),
        ],
    ) -> dict[str, Any]:
        """Search Pixabay for stock images."""
        cfg = ctx.deps.tool_config
        client = get_http_client()

        params = {"q": query, "key": cfg.pixabay_api_key}
        response = await client.get("https://pixabay.com/api/", params=params, follow_redirects=True)
        response.raise_for_status()

        data = response.json()
        if isinstance(data, dict) and "hits" in data:
            await self._validate_results(data["hits"])

        data["system-reminder"] = (
            "All image URLs have been verified for accessibility. "
            "You can use the `download` tool to save the images you need."
        )
        return data

    async def _validate_results(self, results: list[dict[str, Any]]) -> None:
        """Validate image URLs in parallel."""
        semaphore = asyncio.Semaphore(URL_VALIDATION_CONCURRENCY)

        async def validate_one(idx: int) -> None:
            item = results[idx]
            urls = [item.get("webformatURL"), item.get("previewURL"), item.get("largeImageURL")]
            urls = [u for u in urls if u]

            async with semaphore:
                accessible = False
                for url in urls:
                    if await check_url_accessible(url, URL_CHECK_TIMEOUT):
                        accessible = True
                        break

                if not accessible:
                    results[idx] = {
                        "id": item.get("id"),
                        "tags": item.get("tags"),
                        "accessible": False,
                        "unavailable_reason": "Image URLs could not be reached during verification.",
                    }

        tasks = [validate_one(i) for i in range(len(results)) if isinstance(results[i], dict)]
        await asyncio.gather(*tasks)


class SearchImageTool(BaseTool):
    """Real-time image search using RapidAPI."""

    name = "search_image"
    description = "Search real-time images via RapidAPI (similar to Google Images)."

    def is_available(self, ctx: RunContext[AgentContext]) -> bool:
        """Available if RapidAPI key is configured."""
        return bool(ctx.deps.tool_config.rapidapi_api_key)

    async def get_instruction(self, ctx: RunContext[AgentContext]) -> str | None:
        return _load_search_image_instruction()

    async def call(
        self,
        ctx: RunContext[AgentContext],
        query: Annotated[str, Field(description="Search query/keywords")],
        limit: Annotated[int, Field(description="Maximum results to return", default=10)] = 10,
        size: Annotated[
            str,
            Field(description="Image size: any, large, medium, icon, etc.", default="any"),
        ] = "any",
    ) -> dict[str, Any]:
        """Search images via RapidAPI."""
        cfg = ctx.deps.tool_config
        client = get_http_client()

        params = {
            "query": query,
            "limit": limit,
            "size": size,
            "color": "any",
            "type": "any",
            "time": "any",
            "usage_rights": "any",
            "file_type": "any",
            "aspect_ratio": "any",
            "safe_search": "off",
            "region": "us",
        }
        headers = {
            "x-rapidapi-host": "real-time-image-search.p.rapidapi.com",
            "x-rapidapi-key": cfg.rapidapi_api_key,
        }

        response = await client.get(
            "https://real-time-image-search.p.rapidapi.com/search",
            params=params,
            headers=headers,
            follow_redirects=True,
        )
        response.raise_for_status()

        data = response.json()
        if data.get("status") == "ERROR":
            return {"success": False, "error": data.get("message", "Unknown error")}

        if isinstance(data, dict) and "data" in data:
            await self._validate_results(data["data"])

        data["system-reminder"] = (
            "All image URLs have been verified for accessibility. "
            "You can use the `download` tool to save the images you need."
        )
        return data

    async def _validate_results(self, results: list[dict[str, Any]]) -> None:
        """Validate image URLs in parallel."""
        semaphore = asyncio.Semaphore(URL_VALIDATION_CONCURRENCY)

        async def validate_one(idx: int) -> None:
            item = results[idx]
            url = item.get("url")

            if not url:
                results[idx] = self._build_inaccessible(item)
                return

            async with semaphore:
                if not await check_url_accessible(url, URL_CHECK_TIMEOUT):
                    results[idx] = self._build_inaccessible(item)

        tasks = [validate_one(i) for i in range(len(results)) if isinstance(results[i], dict)]
        await asyncio.gather(*tasks)

    def _build_inaccessible(self, item: dict[str, Any]) -> dict[str, Any]:
        """Build response for inaccessible image."""
        return {
            "id": item.get("id"),
            "title": item.get("title"),
            "accessible": False,
            "unavailable_reason": "Image URL could not be reached during verification.",
        }



"""Tests for ya_agent_sdk.events module."""

from dataclasses import dataclass
from datetime import datetime

from ya_agent_sdk.context import AgentContext
from ya_agent_sdk.events import (
    AgentEvent,
    AgentExecutionCompleteEvent,
    AgentExecutionFailedEvent,
    AgentExecutionStartEvent,
    CompactCompleteEvent,
    CompactFailedEvent,
    CompactStartEvent,
    ModelRequestCompleteEvent,
    ModelRequestStartEvent,
    ToolCallsCompleteEvent,
    ToolCallsStartEvent,
)

# =============================================================================
# AgentEvent Tests
# =============================================================================


def test_agent_event_creation() -> None:
    """AgentEvent should be created with event_id and auto timestamp."""
    event = AgentEvent(event_id="test-123")
    assert event.event_id == "test-123"
    assert isinstance(event.timestamp, datetime)


def test_agent_event_custom_timestamp() -> None:
    """AgentEvent should accept custom timestamp."""
    custom_time = datetime(2024, 1, 1, 12, 0, 0)
    event = AgentEvent(event_id="test-123", timestamp=custom_time)
    assert event.timestamp == custom_time


# =============================================================================
# CompactStartEvent Tests
# =============================================================================


def test_compact_start_event_creation() -> None:
    """CompactStartEvent should be created with required fields."""
    event = CompactStartEvent(event_id="compact-001", message_count=50)
    assert event.event_id == "compact-001"
    assert event.message_count == 50
    assert isinstance(event.timestamp, datetime)


def test_compact_start_event_default_message_count() -> None:
    """CompactStartEvent should default message_count to 0."""
    event = CompactStartEvent(event_id="compact-001")
    assert event.message_count == 0


# =============================================================================
# CompactCompleteEvent Tests
# =============================================================================


def test_compact_complete_event_creation() -> None:
    """CompactCompleteEvent should be created with all fields."""
    event = CompactCompleteEvent(
        event_id="compact-001",
        summary_markdown="# Summary\nCompacted content here.",
        original_message_count=100,
        compacted_message_count=5,
    )
    assert event.event_id == "compact-001"
    assert event.summary_markdown == "# Summary\nCompacted content here."
    assert event.original_message_count == 100
    assert event.compacted_message_count == 5


def test_compact_complete_event_defaults() -> None:
    """CompactCompleteEvent should have sensible defaults."""
    event = CompactCompleteEvent(event_id="compact-001")
    assert event.summary_markdown == ""
    assert event.original_message_count == 0
    assert event.compacted_message_count == 0


# =============================================================================
# CompactFailedEvent Tests
# =============================================================================


def test_compact_failed_event_creation() -> None:
    """CompactFailedEvent should be created with error and message_count."""
    event = CompactFailedEvent(
        event_id="compact-001",
        error="API rate limit exceeded",
        message_count=50,
    )
    assert event.event_id == "compact-001"
    assert event.error == "API rate limit exceeded"
    assert event.message_count == 50


def test_compact_failed_event_defaults() -> None:
    """CompactFailedEvent should have sensible defaults."""
    event = CompactFailedEvent(event_id="compact-001")
    assert event.error == ""
    assert event.message_count == 0


# =============================================================================
# AgentContext.emit_event Tests
# =============================================================================


async def test_emit_event_puts_event_in_queue(agent_context: AgentContext) -> None:
    """emit_event should put event into the agent's stream queue."""
    # Enable stream queue for testing
    agent_context._stream_queue_enabled = True

    event = CompactStartEvent(event_id="test-001", message_count=10)

    await agent_context.emit_event(event)

    queue = agent_context.agent_stream_queues[agent_context.agent_id]
    assert not queue.empty()

    received = await queue.get()
    assert received is event
    assert received.event_id == "test-001"
    assert received.message_count == 10


async def test_emit_event_multiple_events(agent_context: AgentContext) -> None:
    """emit_event should queue multiple events in order."""
    # Enable stream queue for testing
    agent_context._stream_queue_enabled = True

    start_event = CompactStartEvent(event_id="test-001", message_count=50)
    complete_event = CompactCompleteEvent(
        event_id="test-001",
        summary_markdown="Done",
        original_message_count=50,
        compacted_message_count=3,
    )

    await agent_context.emit_event(start_event)
    await agent_context.emit_event(complete_event)

    queue = agent_context.agent_stream_queues[agent_context.agent_id]

    first = await queue.get()
    assert isinstance(first, CompactStartEvent)
    assert first.event_id == "test-001"

    second = await queue.get()
    assert isinstance(second, CompactCompleteEvent)
    assert second.event_id == "test-001"


async def test_emit_event_failed_event(agent_context: AgentContext) -> None:
    """emit_event should handle CompactFailedEvent."""
    # Enable stream queue for testing
    agent_context._stream_queue_enabled = True

    start_event = CompactStartEvent(event_id="test-fail", message_count=30)
    failed_event = CompactFailedEvent(event_id="test-fail", error="Model unavailable", message_count=30)

    await agent_context.emit_event(start_event)
    await agent_context.emit_event(failed_event)

    queue = agent_context.agent_stream_queues[agent_context.agent_id]

    first = await queue.get()
    assert isinstance(first, CompactStartEvent)

    second = await queue.get()
    assert isinstance(second, CompactFailedEvent)
    assert second.error == "Model unavailable"
    assert second.message_count == 30


async def test_emit_event_noop_when_disabled(agent_context: AgentContext) -> None:
    """emit_event should be no-op when stream queue is disabled."""
    # Ensure stream queue is disabled (default)
    assert not agent_context._stream_queue_enabled

    event = CompactStartEvent(event_id="test-noop", message_count=10)
    await agent_context.emit_event(event)

    # Queue should remain empty since emit_event is no-op
    queue = agent_context.agent_stream_queues[agent_context.agent_id]
    assert queue.empty()


# =============================================================================
# Event Correlation Tests
# =============================================================================


def test_event_id_correlation() -> None:
    """Start and Complete events should be correlated by event_id."""
    event_id = "correlation-test-abc"

    start = CompactStartEvent(event_id=event_id, message_count=100)
    complete = CompactCompleteEvent(
        event_id=event_id,
        summary_markdown="Summary",
        original_message_count=100,
        compacted_message_count=5,
    )

    assert start.event_id == complete.event_id


def test_event_id_correlation_with_failed() -> None:
    """Start and Failed events should be correlated by event_id."""
    event_id = "fail-correlation-xyz"

    start = CompactStartEvent(event_id=event_id, message_count=80)
    failed = CompactFailedEvent(event_id=event_id, error="Timeout", message_count=80)

    assert start.event_id == failed.event_id


# =============================================================================
# Custom Event Subclass Test
# =============================================================================


def test_custom_event_subclass() -> None:
    """Users should be able to create custom event subclasses."""

    @dataclass
    class CustomAgentEvent(AgentEvent):
        custom_field: str = ""

    event = CustomAgentEvent(event_id="custom-001", custom_field="my value")
    assert event.event_id == "custom-001"
    assert event.custom_field == "my value"
    assert isinstance(event.timestamp, datetime)


# =============================================================================
# Agent Lifecycle Events Tests
# =============================================================================


def test_agent_execution_start_event() -> None:
    """AgentExecutionStartEvent should capture execution start info."""
    event = AgentExecutionStartEvent(
        event_id="exec-001",
        user_prompt="Hello, how are you?",
        message_history_count=5,
    )
    assert event.event_id == "exec-001"
    assert event.user_prompt == "Hello, how are you?"
    assert event.message_history_count == 5
    assert event.deferred_tool_results is None


def test_agent_execution_start_event_defaults() -> None:
    """AgentExecutionStartEvent should have sensible defaults."""
    event = AgentExecutionStartEvent(event_id="exec-001")
    assert event.user_prompt is None
    assert event.deferred_tool_results is None
    assert event.message_history_count == 0


def test_agent_execution_complete_event() -> None:
    """AgentExecutionCompleteEvent should capture execution results."""
    event = AgentExecutionCompleteEvent(
        event_id="exec-001",
        total_loops=3,
        total_duration_seconds=12.5,
        final_message_count=20,
    )
    assert event.event_id == "exec-001"
    assert event.total_loops == 3
    assert event.total_duration_seconds == 12.5
    assert event.final_message_count == 20


def test_agent_execution_complete_event_defaults() -> None:
    """AgentExecutionCompleteEvent should have sensible defaults."""
    event = AgentExecutionCompleteEvent(event_id="exec-001")
    assert event.total_loops == 0
    assert event.total_duration_seconds == 0.0
    assert event.final_message_count == 0


def test_agent_execution_failed_event() -> None:
    """AgentExecutionFailedEvent should capture failure info."""
    event = AgentExecutionFailedEvent(
        event_id="exec-001",
        error="Rate limit exceeded",
        error_type="RateLimitError",
        total_loops=2,
        total_duration_seconds=5.3,
    )
    assert event.event_id == "exec-001"
    assert event.error == "Rate limit exceeded"
    assert event.error_type == "RateLimitError"
    assert event.total_loops == 2
    assert event.total_duration_seconds == 5.3


def test_agent_execution_failed_event_defaults() -> None:
    """AgentExecutionFailedEvent should have sensible defaults."""
    event = AgentExecutionFailedEvent(event_id="exec-001")
    assert event.error == ""
    assert event.error_type == ""
    assert event.total_loops == 0
    assert event.total_duration_seconds == 0.0


# =============================================================================
# Model Request Events Tests
# =============================================================================


def test_model_request_start_event() -> None:
    """ModelRequestStartEvent should capture model request start info."""
    event = ModelRequestStartEvent(
        event_id="exec-001",
        loop_index=2,
        message_count=15,
    )
    assert event.event_id == "exec-001"
    assert event.loop_index == 2
    assert event.message_count == 15


def test_model_request_start_event_defaults() -> None:
    """ModelRequestStartEvent should have sensible defaults."""
    event = ModelRequestStartEvent(event_id="exec-001")
    assert event.loop_index == 0
    assert event.message_count == 0


def test_model_request_complete_event() -> None:
    """ModelRequestCompleteEvent should capture model request completion info."""
    event = ModelRequestCompleteEvent(
        event_id="exec-001",
        loop_index=0,
        duration_seconds=2.5,
    )
    assert event.event_id == "exec-001"
    assert event.loop_index == 0
    assert event.duration_seconds == 2.5


def test_model_request_complete_event_defaults() -> None:
    """ModelRequestCompleteEvent should have sensible defaults."""
    event = ModelRequestCompleteEvent(event_id="exec-001")
    assert event.loop_index == 0
    assert event.duration_seconds == 0.0


# =============================================================================
# Tool Calls Events Tests
# =============================================================================


def test_tool_calls_start_event() -> None:
    """ToolCallsStartEvent should capture tool execution start info."""
    event = ToolCallsStartEvent(
        event_id="exec-001",
        loop_index=1,
    )
    assert event.event_id == "exec-001"
    assert event.loop_index == 1


def test_tool_calls_start_event_defaults() -> None:
    """ToolCallsStartEvent should have sensible defaults."""
    event = ToolCallsStartEvent(event_id="exec-001")
    assert event.loop_index == 0


def test_tool_calls_complete_event() -> None:
    """ToolCallsCompleteEvent should capture tool execution completion info."""
    event = ToolCallsCompleteEvent(
        event_id="exec-001",
        loop_index=0,
        duration_seconds=1.2,
    )
    assert event.event_id == "exec-001"
    assert event.loop_index == 0
    assert event.duration_seconds == 1.2


def test_tool_calls_complete_event_defaults() -> None:
    """ToolCallsCompleteEvent should have sensible defaults."""
    event = ToolCallsCompleteEvent(event_id="exec-001")
    assert event.loop_index == 0
    assert event.duration_seconds == 0.0


# =============================================================================
# Lifecycle Event Correlation Tests
# =============================================================================


def test_lifecycle_event_correlation() -> None:
    """Lifecycle events should be correlated by event_id."""
    event_id = "lifecycle-test-001"

    start = AgentExecutionStartEvent(event_id=event_id, user_prompt="Test")
    model_start = ModelRequestStartEvent(event_id=event_id, loop_index=0)
    model_complete = ModelRequestCompleteEvent(event_id=event_id, loop_index=0, duration_seconds=1.0)
    tool_start = ToolCallsStartEvent(event_id=event_id, loop_index=0)
    tool_complete = ToolCallsCompleteEvent(event_id=event_id, loop_index=0, duration_seconds=0.5)
    complete = AgentExecutionCompleteEvent(event_id=event_id, total_loops=1)

    all_events = [start, model_start, model_complete, tool_start, tool_complete, complete]
    assert all(e.event_id == event_id for e in all_events)



"""Tests for ya_agent_sdk.toolsets.core.filesystem.view module."""

from contextlib import AsyncExitStack
from pathlib import Path
from unittest.mock import MagicMock, patch

from inline_snapshot import snapshot
from pydantic_ai import BinaryContent, RunContext, ToolReturn

from ya_agent_sdk.context import AgentContext
from ya_agent_sdk.environment.local import LocalEnvironment
from ya_agent_sdk.toolsets.core.filesystem.view import (
    IMAGE_EXTENSIONS,
    MEDIA_TYPE_MAP,
    SUPPORTED_IMAGE_MEDIA_TYPES,
    VIDEO_EXTENSIONS,
    ViewTool,
)


async def test_view_tool_attributes(agent_context: AgentContext) -> None:
    """Should have correct name and description."""
    assert ViewTool.name == "view"
    assert "Read files" in ViewTool.description
    tool = ViewTool()
    mock_run_ctx = MagicMock(spec=RunContext)
    mock_run_ctx.deps = agent_context
    instruction = await tool.get_instruction(mock_run_ctx)
    assert instruction is not None


def test_view_tool_initialization(agent_context: AgentContext) -> None:
    """Should initialize with context."""
    tool = ViewTool()
    assert tool.name == "view"


def test_view_tool_is_available(agent_context: AgentContext, mock_run_ctx) -> None:
    """Should be available by default."""
    tool = ViewTool()
    assert tool.is_available(mock_run_ctx) is True


def test_is_image_file(agent_context: AgentContext) -> None:
    """Should correctly identify image files."""
    tool = ViewTool()
    for ext in IMAGE_EXTENSIONS:
        assert tool._is_image_file(f"test{ext}") is True
        assert tool._is_image_file(f"test{ext.upper()}") is True
    assert tool._is_image_file("test.txt") is False
    assert tool._is_image_file("test.py") is False


def test_is_video_file(agent_context: AgentContext) -> None:
    """Should correctly identify video files."""
    tool = ViewTool()
    for ext in VIDEO_EXTENSIONS:
        assert tool._is_video_file(f"test{ext}") is True
        assert tool._is_video_file(f"test{ext.upper()}") is True
    assert tool._is_video_file("test.txt") is False
    assert tool._is_video_file("test.png") is False


def test_get_media_type(agent_context: AgentContext) -> None:
    """Should return correct media type for extensions."""
    tool = ViewTool()
    for ext, expected in MEDIA_TYPE_MAP.items():
        assert tool._get_media_type(f"test{ext}") == expected
    assert tool._get_media_type("test.unknown") == "application/octet-stream"


async def test_view_text_file_simple(tmp_path: Path) -> None:
    """Should read text file and return content string when no truncation."""
    async with AsyncExitStack() as stack:
        env = await stack.enter_async_context(
            LocalEnvironment(allowed_paths=[tmp_path], default_path=tmp_path, tmp_base_dir=tmp_path)
        )
        ctx = await stack.enter_async_context(AgentContext(env=env))
        tool = ViewTool()

        # Create test file
        test_file = tmp_path / "test.txt"
        test_file.write_text("Hello, World!\nLine 2\nLine 3")

        mock_run_ctx = MagicMock(spec=RunContext)
        mock_run_ctx.deps = ctx

        result = await tool.call(mock_run_ctx, file_path="test.txt")
        assert result == "Hello, World!\nLine 2\nLine 3"


async def test_view_text_file_with_offset(tmp_path: Path) -> None:
    """Should read text file with line offset and return metadata."""
    async with AsyncExitStack() as stack:
        env = await stack.enter_async_context(
            LocalEnvironment(allowed_paths=[tmp_path], default_path=tmp_path, tmp_base_dir=tmp_path)
        )
        ctx = await stack.enter_async_context(AgentContext(env=env))
        tool = ViewTool()

        # Create test file with multiple lines
        lines = [f"Line {i}" for i in range(10)]
        test_file = tmp_path / "test.txt"
        test_file.write_text("\n".join(lines))

        mock_run_ctx = MagicMock(spec=RunContext)
        mock_run_ctx.deps = ctx

        result = await tool.call(mock_run_ctx, file_path="test.txt", line_offset=5)
        assert isinstance(result, dict)
        assert "content" in result
        assert "metadata" in result
        assert "Line 5" in result["content"]
        assert result["metadata"]["current_segment"]["start_line"] == 6


async def test_view_text_file_with_limit(tmp_path: Path) -> None:
    """Should truncate content when exceeding line limit."""
    async with AsyncExitStack() as stack:
        env = await stack.enter_async_context(
            LocalEnvironment(allowed_paths=[tmp_path], default_path=tmp_path, tmp_base_dir=tmp_path)
        )
        ctx = await stack.enter_async_context(AgentContext(env=env))
        tool = ViewTool()

        # Create file with many lines
        lines = [f"Line {i}" for i in range(100)]
        test_file = tmp_path / "test.txt"
        test_file.write_text("\n".join(lines))

        mock_run_ctx = MagicMock(spec=RunContext)
        mock_run_ctx.deps = ctx

        result = await tool.call(mock_run_ctx, file_path="test.txt", line_limit=10)
        assert isinstance(result, dict)
        assert result["metadata"]["current_segment"]["lines_to_show"] == 10
        assert result["metadata"]["current_segment"]["has_more_content"] is True


async def test_view_text_file_line_truncation(tmp_path: Path) -> None:
    """Should truncate long lines."""
    async with AsyncExitStack() as stack:
        env = await stack.enter_async_context(
            LocalEnvironment(allowed_paths=[tmp_path], default_path=tmp_path, tmp_base_dir=tmp_path)
        )
        ctx = await stack.enter_async_context(AgentContext(env=env))
        tool = ViewTool()

        # Create file with long line
        long_line = "A" * 3000
        test_file = tmp_path / "test.txt"
        test_file.write_text(long_line)

        mock_run_ctx = MagicMock(spec=RunContext)
        mock_run_ctx.deps = ctx

        result = await tool.call(mock_run_ctx, file_path="test.txt", max_line_length=100)
        assert isinstance(result, dict)
        assert "(line truncated)" in result["content"]
        assert result["metadata"]["truncation_info"]["lines_truncated"] is True


async def test_view_file_not_found(tmp_path: Path) -> None:
    """Should return error message when file not found."""
    async with AsyncExitStack() as stack:
        env = await stack.enter_async_context(
            LocalEnvironment(allowed_paths=[tmp_path], default_path=tmp_path, tmp_base_dir=tmp_path)
        )
        ctx = await stack.enter_async_context(AgentContext(env=env))
        tool = ViewTool()

        mock_run_ctx = MagicMock(spec=RunContext)
        mock_run_ctx.deps = ctx

        result = await tool.call(mock_run_ctx, file_path="nonexistent.txt")
        assert result == snapshot("Error: File not found: nonexistent.txt")


async def test_view_directory_error(tmp_path: Path) -> None:
    """Should return error when path is a directory."""
    async with AsyncExitStack() as stack:
        env = await stack.enter_async_context(
            LocalEnvironment(allowed_paths=[tmp_path], default_path=tmp_path, tmp_base_dir=tmp_path)
        )
        ctx = await stack.enter_async_context(AgentContext(env=env))
        tool = ViewTool()

        # Create a directory
        test_dir = tmp_path / "testdir"
        test_dir.mkdir()

        mock_run_ctx = MagicMock(spec=RunContext)
        mock_run_ctx.deps = ctx

        result = await tool.call(mock_run_ctx, file_path="testdir")
        assert result == snapshot("Error: Path is a directory, not a file: testdir")


async def test_view_image_file(tmp_path: Path) -> None:
    """Should return ToolReturn with BinaryContent for image files."""
    from ya_agent_sdk.context import ModelCapability, ModelConfig

    async with AsyncExitStack() as stack:
        env = await stack.enter_async_context(
            LocalEnvironment(allowed_paths=[tmp_path], default_path=tmp_path, tmp_base_dir=tmp_path)
        )
        # Create context with vision capability
        ctx = await stack.enter_async_context(
            AgentContext(
                env=env,
                model_cfg=ModelConfig(capabilities={ModelCapability.vision}),
            )
        )
        tool = ViewTool()

        # Create a minimal PNG file (1x1 transparent pixel)
        png_data = bytes([
            0x89,
            0x50,
            0x4E,
            0x47,
            0x0D,
            0x0A,
            0x1A,
            0x0A,
            0x00,
            0x00,
            0x00,
            0x0D,
            0x49,
            0x48,
            0x44,
            0x52,
            0x00,
            0x00,
            0x00,
            0x01,
            0x00,
            0x00,
            0x00,
            0x01,
            0x08,
            0x06,
            0x00,
            0x00,
            0x00,
            0x1F,
            0x15,
            0xC4,
            0x89,
            0x00,
            0x00,
            0x00,
            0x0A,
            0x49,
            0x44,
            0x41,
            0x54,
            0x78,
            0x9C,
            0x63,
            0x00,
            0x01,
            0x00,
            0x00,
            0x05,
            0x00,
            0x01,
            0x0D,
            0x0A,
            0x2D,
            0xB4,
            0x00,
            0x00,
            0x00,
            0x00,
            0x49,
            0x45,
            0x4E,
            0x44,
            0xAE,
            0x42,
            0x60,
            0x82,
        ])
        test_file = tmp_path / "test.png"
        test_file.write_bytes(png_data)

        mock_run_ctx = MagicMock(spec=RunContext)
        mock_run_ctx.deps = ctx

        result = await tool.call(mock_run_ctx, file_path="test.png")
        assert isinstance(result, ToolReturn)
        assert "image is attached" in result.return_value
        assert len(result.content) == 1
        assert isinstance(result.content[0], BinaryContent)
        assert result.content[0].media_type == "image/png"


async def test_view_reject_large_image_inline(tmp_path: Path) -> None:
    """Should reject oversized images before loading them into memory."""
    from ya_agent_sdk.context import ModelCapability, ModelConfig

    async with AsyncExitStack() as stack:
        env = await stack.enter_async_context(
            LocalEnvironment(allowed_paths=[tmp_path], default_path=tmp_path, tmp_base_dir=tmp_path)
        )
        ctx = await stack.enter_async_context(
            AgentContext(
                env=env,
                model_cfg=ModelConfig(capabilities={ModelCapability.vision}),
            )
        )
        tool = ViewTool()

        test_file = tmp_path / "huge.png"
        test_file.write_bytes(b"x" * (21 * 1024 * 1024))

        mock_run_ctx = MagicMock(spec=RunContext)
        mock_run_ctx.deps = ctx

        result = await tool.call(mock_run_ctx, file_path="huge.png")
        assert result == snapshot(
            "Error: Image file is too large to inline (21.00 MB). Maximum supported inline size is 20.00 MB."
        )


async def test_view_reject_large_text_file(tmp_path: Path) -> None:
    """Should reject text files that exceed safe inspection size."""
    async with AsyncExitStack() as stack:
        env = await stack.enter_async_context(
            LocalEnvironment(allowed_paths=[tmp_path], default_path=tmp_path, tmp_base_dir=tmp_path)
        )
        ctx = await stack.enter_async_context(AgentContext(env=env))
        tool = ViewTool()

        test_file = tmp_path / "huge.txt"
        test_file.write_text("a" * (11 * 1024 * 1024), encoding="utf-8")

        mock_run_ctx = MagicMock(spec=RunContext)
        mock_run_ctx.deps = ctx

        result = await tool.call(mock_run_ctx, file_path="huge.txt")
        assert result == snapshot({
            "error": "File is too large to inspect safely (11.00 MB). Maximum supported text view size is 10.00 MB. Use shell tools (e.g. `head`, `tail`, `sed -n`) to read portions of this file.",
            "success": False,
        })


async def test_view_uses_tool_config_text_limit(tmp_path: Path) -> None:
    """Should honor custom text size limit from ToolConfig."""
    from ya_agent_sdk.context import ToolConfig

    async with AsyncExitStack() as stack:
        env = await stack.enter_async_context(
            LocalEnvironment(allowed_paths=[tmp_path], default_path=tmp_path, tmp_base_dir=tmp_path)
        )
        ctx = await stack.enter_async_context(
            AgentContext(
                env=env,
                tool_config=ToolConfig(view_max_text_file_size=1024),
            )
        )
        tool = ViewTool()

        test_file = tmp_path / "custom-limit.txt"
        test_file.write_text("a" * 2048, encoding="utf-8")

        mock_run_ctx = MagicMock(spec=RunContext)
        mock_run_ctx.deps = ctx

        result = await tool.call(mock_run_ctx, file_path="custom-limit.txt")
        assert result == snapshot({
            "error": "File is too large to inspect safely (2.0 KB). Maximum supported text view size is 1.0 KB. Use shell tools (e.g. `head`, `tail`, `sed -n`) to read portions of this file.",
            "success": False,
        })


async def test_view_video_file_with_video_model(tmp_path: Path) -> None:
    """Should return video content when model supports video."""
    from ya_agent_sdk.context import ModelCapability, ModelConfig

    async with AsyncExitStack() as stack:
        env = await stack.enter_async_context(
            LocalEnvironment(allowed_paths=[tmp_path], default_path=tmp_path, tmp_base_dir=tmp_path)
        )
        # Create context with video_understanding capability
        ctx = await stack.enter_async_context(
            AgentContext(
                env=env,
                model_cfg=ModelConfig(capabilities={ModelCapability.video_understanding}),
            )
        )
        tool = ViewTool()

        # Create a minimal video file
        test_file = tmp_path / "test.mp4"
        test_file.write_bytes(b"fake video data")

        mock_run_ctx = MagicMock(spec=RunContext)
        mock_run_ctx.deps = ctx
        mock_run_ctx.tool_call_id = "test-id"

        result = await tool.call(mock_run_ctx, file_path="test.mp4")

        assert isinstance(result, ToolReturn)
        assert "video is attached" in result.return_value
        assert len(result.content) == 1
        assert result.content[0].media_type == "video/mp4"


async def test_view_video_file_fallback_to_image_understanding(tmp_path: Path) -> None:
    """Should fallback to image understanding when model doesn't support video."""
    async with AsyncExitStack() as stack:
        env = await stack.enter_async_context(
            LocalEnvironment(allowed_paths=[tmp_path], default_path=tmp_path, tmp_base_dir=tmp_path)
        )
        ctx = await stack.enter_async_context(AgentContext(env=env))
        tool = ViewTool()

        # Create a minimal video file
        test_file = tmp_path / "test.mp4"
        test_file.write_bytes(b"fake video data")

        mock_run_ctx = MagicMock(spec=RunContext)
        mock_run_ctx.deps = ctx
        mock_run_ctx.tool_call_id = "test-id"

        # Mock _read_video_with_fallback to simulate fallback behavior
        async def mock_fallback(path, file_path, run_ctx):
            return "Video description (via image analysis):\nThis video shows a test scene."

        with patch.object(tool, "_read_video_with_fallback", side_effect=mock_fallback):
            result = await tool.call(mock_run_ctx, file_path="test.mp4")
            assert "Video description" in result
            assert "test scene" in result


async def test_view_video_fallback_failure(tmp_path: Path) -> None:
    """Should return error message when video fallback fails."""
    async with AsyncExitStack() as stack:
        env = await stack.enter_async_context(
            LocalEnvironment(allowed_paths=[tmp_path], default_path=tmp_path, tmp_base_dir=tmp_path)
        )
        ctx = await stack.enter_async_context(AgentContext(env=env))
        tool = ViewTool()

        # Create a minimal video file
        test_file = tmp_path / "test.mp4"
        test_file.write_bytes(b"fake video data")

        mock_run_ctx = MagicMock(spec=RunContext)
        mock_run_ctx.deps = ctx
        mock_run_ctx.tool_call_id = "test-id"

        # Mock _read_video_with_fallback to simulate fallback failure
        async def mock_fallback_failure(path, file_path, run_ctx):
            return f"Video file: {file_path}. Model does not support video understanding and fallback analysis failed."

        with patch.object(tool, "_read_video_with_fallback", side_effect=mock_fallback_failure):
            result = await tool.call(mock_run_ctx, file_path="test.mp4")
            assert "does not support video understanding" in result


async def test_view_webm_video(tmp_path: Path) -> None:
    """Should handle webm video with correct media type."""
    from ya_agent_sdk.context import ModelCapability, ModelConfig

    async with AsyncExitStack() as stack:
        env = await stack.enter_async_context(
            LocalEnvironment(allowed_paths=[tmp_path], default_path=tmp_path, tmp_base_dir=tmp_path)
        )
        # Create context with video_understanding capability
        ctx = await stack.enter_async_context(
            AgentContext(
                env=env,
                model_cfg=ModelConfig(capabilities={ModelCapability.video_understanding}),
            )
        )
        tool = ViewTool()

        test_file = tmp_path / "test.webm"
        test_file.write_bytes(b"fake webm data")

        mock_run_ctx = MagicMock(spec=RunContext)
        mock_run_ctx.deps = ctx
        mock_run_ctx.tool_call_id = "test-id"

        result = await tool.call(mock_run_ctx, file_path="test.webm")

        assert isinstance(result, ToolReturn)
        assert result.content[0].media_type == "video/webm"


def test_supported_image_media_types() -> None:
    """Should have expected supported image media types."""
    assert "image/png" in SUPPORTED_IMAGE_MEDIA_TYPES
    assert "image/jpeg" in SUPPORTED_IMAGE_MEDIA_TYPES
    assert "image/webp" in SUPPORTED_IMAGE_MEDIA_TYPES



"""Fixtures for toolset tests."""

from unittest.mock import MagicMock

import pytest
from pydantic_ai import RunContext

from ya_agent_sdk.context import AgentContext


@pytest.fixture
def mock_ctx() -> AgentContext:
    """Create a mock AgentContext for testing."""
    return AgentContext(run_id="test-run-123")


@pytest.fixture
def mock_run_context(mock_ctx: AgentContext) -> RunContext[AgentContext]:
    """Create a mock RunContext."""
    ctx = MagicMock(spec=RunContext)
    ctx.deps = mock_ctx
    return ctx



"""Tests for configuration management using pydantic-settings."""

from ya_agent_sdk.toolsets.browser_use._config import BrowserUseSettings
from ya_agent_sdk.toolsets.browser_use.toolset import BrowserUseToolset


def test_browser_use_settings_defaults():
    """Test that BrowserUseSettings has correct default values."""
    settings = BrowserUseSettings()
    assert settings.max_retries == 3
    assert settings.prefix is None
    assert settings.always_use_new_page is False
    assert settings.auto_cleanup_page is False


def test_browser_use_settings_from_env(monkeypatch):
    """Test that BrowserUseSettings loads from environment variables."""
    # Set environment variables
    monkeypatch.setenv("YA_AGENT_BROWSER_USE_MAX_RETRIES", "5")
    monkeypatch.setenv("YA_AGENT_BROWSER_USE_PREFIX", "custom_prefix")
    monkeypatch.setenv("YA_AGENT_BROWSER_USE_ALWAYS_USE_NEW_PAGE", "true")
    monkeypatch.setenv("YA_AGENT_BROWSER_USE_AUTO_CLEANUP_PAGE", "false")

    settings = BrowserUseSettings()
    assert settings.max_retries == 5
    assert settings.prefix == "custom_prefix"
    assert settings.always_use_new_page is True
    assert settings.auto_cleanup_page is False


def test_browser_use_toolset_uses_settings_defaults():
    """Test that BrowserUseToolset uses fallback defaults when settings and parameters are None."""
    toolset = BrowserUseToolset(cdp_url="http://localhost:9222/json/version")
    assert toolset.max_retries == 3  # Default fallback
    assert toolset.prefix == "browser_use"  # Default to toolset.id
    assert toolset.always_use_new_page is False  # Default fallback
    assert toolset.auto_cleanup_page is False  # Default is False


def test_browser_use_toolset_uses_env_settings(monkeypatch):
    """Test that BrowserUseToolset uses environment variables when parameters are None."""
    # Set environment variables
    monkeypatch.setenv("YA_AGENT_BROWSER_USE_MAX_RETRIES", "10")
    monkeypatch.setenv("YA_AGENT_BROWSER_USE_PREFIX", "env_browser")
    monkeypatch.setenv("YA_AGENT_BROWSER_USE_ALWAYS_USE_NEW_PAGE", "true")

    toolset = BrowserUseToolset(cdp_url="http://localhost:9222/json/version")
    assert toolset.max_retries == 10
    assert toolset.prefix == "env_browser"
    assert toolset.always_use_new_page is True
    assert toolset.auto_cleanup_page is False  # Default is False regardless of always_use_new_page


def test_browser_use_toolset_parameters_override_env(monkeypatch):
    """Test that explicit parameters override environment variables."""
    # Set environment variables
    monkeypatch.setenv("YA_AGENT_BROWSER_USE_MAX_RETRIES", "10")
    monkeypatch.setenv("YA_AGENT_BROWSER_USE_PREFIX", "env_browser")
    monkeypatch.setenv("YA_AGENT_BROWSER_USE_ALWAYS_USE_NEW_PAGE", "true")
    monkeypatch.setenv("YA_AGENT_BROWSER_USE_AUTO_CLEANUP_PAGE", "true")

    # Create toolset with explicit parameters
    toolset = BrowserUseToolset(
        cdp_url="http://localhost:9222/json/version",
        max_retries=20,
        prefix="param_browser",
        always_use_new_page=False,
        auto_cleanup_page=False,
    )

    # Parameters should override environment variables
    assert toolset.max_retries == 20
    assert toolset.prefix == "param_browser"
    assert toolset.always_use_new_page is False
    assert toolset.auto_cleanup_page is False


def test_browser_use_toolset_partial_parameter_override(monkeypatch):
    """Test that only provided parameters override environment variables."""
    # Set environment variables
    monkeypatch.setenv("YA_AGENT_BROWSER_USE_MAX_RETRIES", "10")
    monkeypatch.setenv("YA_AGENT_BROWSER_USE_PREFIX", "env_browser")
    monkeypatch.setenv("YA_AGENT_BROWSER_USE_ALWAYS_USE_NEW_PAGE", "true")
    monkeypatch.setenv("YA_AGENT_BROWSER_USE_AUTO_CLEANUP_PAGE", "false")

    # Create toolset with only max_retries parameter
    toolset = BrowserUseToolset(
        cdp_url="http://localhost:9222/json/version",
        max_retries=20,  # Override env
        # prefix, always_use_new_page and auto_cleanup_page will use env values
    )

    assert toolset.max_retries == 20  # Overridden by parameter
    assert toolset.prefix == "env_browser"  # From environment
    assert toolset.always_use_new_page is True  # From environment
    assert toolset.auto_cleanup_page is False  # From environment


def test_browser_use_settings_env_file_loading(tmp_path, monkeypatch):
    """Test that BrowserUseSettings can load from .env file."""
    # Create a temporary .env file
    env_file = tmp_path / ".env"
    env_file.write_text(
        "YA_AGENT_BROWSER_USE_MAX_RETRIES=15\n"
        "YA_AGENT_BROWSER_USE_PREFIX=file_browser\n"
        "YA_AGENT_BROWSER_USE_ALWAYS_USE_NEW_PAGE=false\n"
        "YA_AGENT_BROWSER_USE_AUTO_CLEANUP_PAGE=true\n"
    )

    # Change to temp directory to pick up .env file
    monkeypatch.chdir(tmp_path)

    settings = BrowserUseSettings()
    assert settings.max_retries == 15
    assert settings.prefix == "file_browser"
    assert settings.always_use_new_page is False
    assert settings.auto_cleanup_page is True


def test_auto_cleanup_page_defaults_to_false():
    """Test that auto_cleanup_page defaults to False regardless of always_use_new_page."""
    # When always_use_new_page is True, auto_cleanup_page should still default to False
    toolset = BrowserUseToolset(
        cdp_url="http://localhost:9222/json/version",
        always_use_new_page=True,
    )
    assert toolset.auto_cleanup_page is False

    # When always_use_new_page is False, auto_cleanup_page should also default to False
    toolset = BrowserUseToolset(
        cdp_url="http://localhost:9222/json/version",
        always_use_new_page=False,
    )
    assert toolset.auto_cleanup_page is False

    # Default behavior without any parameters
    toolset = BrowserUseToolset(
        cdp_url="http://localhost:9222/json/version",
    )
    assert toolset.auto_cleanup_page is False


def test_auto_cleanup_page_can_override_default():
    """Test that auto_cleanup_page parameter can override the default behavior."""
    # Explicitly disable auto_cleanup even when always_use_new_page is True
    toolset = BrowserUseToolset(
        cdp_url="http://localhost:9222/json/version",
        always_use_new_page=True,
        auto_cleanup_page=False,
    )
    assert toolset.always_use_new_page is True
    assert toolset.auto_cleanup_page is False

    # Explicitly enable auto_cleanup even when always_use_new_page is False
    toolset = BrowserUseToolset(
        cdp_url="http://localhost:9222/json/version",
        always_use_new_page=False,
        auto_cleanup_page=True,
    )
    assert toolset.always_use_new_page is False
    assert toolset.auto_cleanup_page is True



"""Tests for ToolSearchToolSet, ToolMetadata, and search strategies (namespace-aware)."""

from __future__ import annotations

from typing import Annotated
from unittest.mock import MagicMock

import pytest
from pydantic import Field
from pydantic_ai import RunContext

from ya_agent_sdk.context import AgentContext
from ya_agent_sdk.toolsets.core.base import BaseTool, Toolset
from ya_agent_sdk.toolsets.tool_search.metadata import ToolMetadata, extract_metadata_from_schema
from ya_agent_sdk.toolsets.tool_search.strategies.keyword import KeywordSearchStrategy
from ya_agent_sdk.toolsets.tool_search.toolset import ToolSearchToolSet

# ---------------------------------------------------------------------------
# Test tools
# ---------------------------------------------------------------------------


class GetWeatherTool(BaseTool):
    name = "get_weather"
    description = "Get the current weather in a given location"

    async def call(
        self,
        ctx: RunContext[AgentContext],
        location: Annotated[str, Field(description="The city and state")],
        unit: Annotated[str, Field(description="Temperature unit", default="celsius")] = "celsius",
    ) -> str:
        return f"Weather in {location}: sunny, 25{unit[0]}"


class GetForecastTool(BaseTool):
    name = "get_forecast"
    description = "Get the weather forecast for multiple days ahead"

    async def call(
        self,
        ctx: RunContext[AgentContext],
        location: Annotated[str, Field(description="The city name")],
        days: Annotated[int, Field(description="Number of days to forecast")] = 5,
    ) -> str:
        return f"Forecast for {location}: {days} days"


class GetStockPriceTool(BaseTool):
    name = "get_stock_price"
    description = "Get the current stock price for a ticker symbol"

    async def call(
        self,
        ctx: RunContext[AgentContext],
        ticker: Annotated[str, Field(description="Stock ticker symbol like AAPL")],
    ) -> str:
        return f"Stock {ticker}: $150.00"


class ConvertCurrencyTool(BaseTool):
    name = "convert_currency"
    description = "Convert an amount from one currency to another"

    async def call(
        self,
        ctx: RunContext[AgentContext],
        amount: Annotated[float, Field(description="Amount to convert")],
        from_currency: Annotated[str, Field(description="Source currency code")],
        to_currency: Annotated[str, Field(description="Target currency code")],
    ) -> str:
        return f"{amount} {from_currency} = {amount * 0.85} {to_currency}"


class ViewFileTool(BaseTool):
    name = "view"
    description = "Read contents of a file from the filesystem"

    async def call(
        self,
        ctx: RunContext[AgentContext],
        file_path: Annotated[str, Field(description="Path to the file to read")],
    ) -> str:
        return f"Contents of {file_path}"


class EditFileTool(BaseTool):
    name = "edit"
    description = "Edit a file by replacing text"

    async def call(
        self,
        ctx: RunContext[AgentContext],
        file_path: Annotated[str, Field(description="Path to the file")],
        old_string: Annotated[str, Field(description="Text to replace")],
        new_string: Annotated[str, Field(description="Replacement text")],
    ) -> str:
        return f"Edited {file_path}"


# ---------------------------------------------------------------------------
# Fixtures
# ---------------------------------------------------------------------------


@pytest.fixture
def mock_ctx() -> AgentContext:
    return AgentContext(run_id="test-tool-search")


@pytest.fixture
def mock_run_context(mock_ctx: AgentContext) -> RunContext[AgentContext]:
    ctx = MagicMock(spec=RunContext)
    ctx.deps = mock_ctx
    return ctx


@pytest.fixture
def weather_toolset() -> Toolset:
    """Weather toolset with namespace id."""
    return Toolset(tools=[GetWeatherTool, GetForecastTool], toolset_id="weather")


@pytest.fixture
def finance_toolset() -> Toolset:
    """Finance toolset with namespace id."""
    return Toolset(tools=[GetStockPriceTool, ConvertCurrencyTool], toolset_id="finance")


@pytest.fixture
def loose_toolset() -> Toolset:
    """Loose toolset without id (individual tool loading)."""
    return Toolset(tools=[ViewFileTool, EditFileTool])


@pytest.fixture
def weather_toolset_no_id() -> Toolset:
    """Weather toolset without namespace id (loose)."""
    return Toolset(tools=[GetWeatherTool, GetForecastTool])


# ---------------------------------------------------------------------------
# ToolMetadata tests
# ---------------------------------------------------------------------------


def test_tool_metadata_searchable_text():
    meta = ToolMetadata(
        name="get_weather",
        description="Get current weather",
        parameter_names=["location", "unit"],
        parameter_descriptions={"location": "City name", "unit": "Temperature unit"},
    )
    text = meta.searchable_text
    assert "get_weather" in text
    assert "Get current weather" in text
    assert "location" in text
    assert "City name" in text
    assert "unit" in text


def test_tool_metadata_searchable_text_with_namespace():
    meta = ToolMetadata(
        name="get_weather",
        description="Get current weather",
        namespace="weather",
    )
    assert "Namespace: weather" in meta.searchable_text


def test_tool_metadata_namespace_entry_searchable_text():
    meta = ToolMetadata(
        name="weather",
        description="Weather related tools",
        is_namespace_entry=True,
        namespace="weather",
        namespace_tool_names=["get_weather", "get_forecast"],
    )
    text = meta.searchable_text
    assert "Namespace: weather" in text
    assert "Weather related tools" in text
    assert "get_weather" in text
    assert "get_forecast" in text


def test_tool_metadata_brief():
    meta = ToolMetadata(
        name="get_weather",
        description="Get current weather",
        parameter_names=["location", "unit"],
    )
    brief = meta.brief
    assert "get_weather" in brief
    assert "location, unit" in brief


def test_tool_metadata_brief_no_params():
    meta = ToolMetadata(name="ping", description="Ping the server")
    assert "none" in meta.brief


def test_tool_metadata_namespace_entry_brief():
    meta = ToolMetadata(
        name="weather",
        description="Weather tools",
        is_namespace_entry=True,
        namespace_tool_names=["get_weather", "get_forecast"],
    )
    brief = meta.brief
    assert "[weather]" in brief
    assert "get_weather" in brief
    assert "get_forecast" in brief


def test_extract_metadata_from_schema():
    schema = {
        "type": "object",
        "properties": {
            "location": {"type": "string", "description": "The city name"},
            "unit": {"type": "string", "description": "Temperature unit"},
        },
        "required": ["location"],
    }
    meta = extract_metadata_from_schema(
        name="get_weather",
        description="Get current weather",
        parameters_json_schema=schema,
        namespace="weather",
    )
    assert meta.name == "get_weather"
    assert meta.description == "Get current weather"
    assert "location" in meta.parameter_names
    assert "unit" in meta.parameter_names
    assert meta.parameter_descriptions["location"] == "The city name"
    assert meta.namespace == "weather"


def test_extract_metadata_empty_schema():
    meta = extract_metadata_from_schema(
        name="ping",
        description=None,
        parameters_json_schema={"type": "object", "properties": {}},
    )
    assert meta.name == "ping"
    assert meta.description == ""
    assert meta.parameter_names == []
    assert meta.namespace is None


# ---------------------------------------------------------------------------
# KeywordSearchStrategy tests
# ---------------------------------------------------------------------------


@pytest.mark.anyio
async def test_keyword_search_basic():
    strategy = KeywordSearchStrategy()
    candidates = [
        ToolMetadata(name="get_weather", description="Get current weather for a location"),
        ToolMetadata(name="get_stock_price", description="Get stock price for a ticker"),
        ToolMetadata(name="convert_currency", description="Convert between currencies"),
    ]
    results = await strategy.search("weather", candidates)
    assert len(results) >= 1
    assert results[0].name == "get_weather"


@pytest.mark.anyio
async def test_keyword_search_regex():
    strategy = KeywordSearchStrategy()
    candidates = [
        ToolMetadata(name="get_weather", description="Get current weather"),
        ToolMetadata(name="get_forecast", description="Get weather forecast"),
        ToolMetadata(name="get_stock_price", description="Get stock price"),
    ]
    results = await strategy.search("get_.*cast", candidates)
    assert len(results) == 1
    assert results[0].name == "get_forecast"


@pytest.mark.anyio
async def test_keyword_search_invalid_regex_fallback():
    strategy = KeywordSearchStrategy()
    candidates = [
        ToolMetadata(name="get_weather", description="Get current weather"),
        ToolMetadata(name="test[tool", description="A tool with brackets in name"),
    ]
    results = await strategy.search("test[tool", candidates)
    assert len(results) == 1
    assert results[0].name == "test[tool"


@pytest.mark.anyio
async def test_keyword_search_scoring_name_beats_description():
    strategy = KeywordSearchStrategy()
    candidates = [
        ToolMetadata(name="weather_tool", description="Some tool"),
        ToolMetadata(name="other_tool", description="Does weather stuff"),
    ]
    results = await strategy.search("weather", candidates)
    assert len(results) == 2
    assert results[0].name == "weather_tool"


@pytest.mark.anyio
async def test_keyword_search_max_results():
    strategy = KeywordSearchStrategy()
    candidates = [ToolMetadata(name=f"tool_{i}", description="Generic tool") for i in range(10)]
    results = await strategy.search("tool", candidates, max_results=3)
    assert len(results) == 3


@pytest.mark.anyio
async def test_keyword_search_no_match():
    strategy = KeywordSearchStrategy()
    candidates = [ToolMetadata(name="get_weather", description="Get weather")]
    results = await strategy.search("database", candidates)
    assert len(results) == 0


@pytest.mark.anyio
async def test_keyword_search_empty_query():
    strategy = KeywordSearchStrategy()
    candidates = [ToolMetadata(name="tool", description="A tool")]
    assert await strategy.search("", candidates) == []


@pytest.mark.anyio
async def test_keyword_search_empty_candidates():
    strategy = KeywordSearchStrategy()
    assert await strategy.search("weather", []) == []


@pytest.mark.anyio
async def test_keyword_search_parameter_match():
    strategy = KeywordSearchStrategy()
    candidates = [
        ToolMetadata(
            name="send_email",
            description="Send an email message",
            parameter_names=["recipient", "subject", "body"],
            parameter_descriptions={"recipient": "Email address of the recipient"},
        ),
        ToolMetadata(name="get_weather", description="Get weather"),
    ]
    results = await strategy.search("recipient", candidates)
    assert len(results) == 1
    assert results[0].name == "send_email"


@pytest.mark.anyio
async def test_keyword_search_namespace_match():
    strategy = KeywordSearchStrategy()
    candidates = [
        ToolMetadata(name="list_orders", description="List orders", namespace="crm"),
        ToolMetadata(name="get_weather", description="Get weather", namespace="weather"),
    ]
    results = await strategy.search("crm", candidates)
    assert len(results) == 1
    assert results[0].name == "list_orders"


@pytest.mark.anyio
async def test_keyword_search_namespace_entry():
    strategy = KeywordSearchStrategy()
    candidates = [
        ToolMetadata(
            name="weather",
            description="Weather related tools",
            is_namespace_entry=True,
            namespace="weather",
            namespace_tool_names=["get_weather", "get_forecast"],
        ),
        ToolMetadata(name="get_stock_price", description="Get stock price"),
    ]
    results = await strategy.search("weather", candidates)
    assert len(results) >= 1
    # Namespace entry should match (name match + description match)
    ns_results = [r for r in results if r.is_namespace_entry]
    assert len(ns_results) == 1
    assert ns_results[0].name == "weather"


# ---------------------------------------------------------------------------
# ToolSearchToolSet tests - namespace loading
# ---------------------------------------------------------------------------


@pytest.mark.anyio
async def test_toolset_all_deferred_initially(weather_toolset, mock_run_context):
    """With no prior state, only tool_search should be visible."""
    ts = ToolSearchToolSet(toolsets=[weather_toolset])
    tools = await ts.get_tools(mock_run_context)

    assert "tool_search" in tools
    assert "get_weather" not in tools
    assert "get_forecast" not in tools


@pytest.mark.anyio
async def test_tool_search_always_first_in_tool_list(weather_toolset, loose_toolset, mock_run_context):
    """tool_search must always be the first tool for stable positioning."""
    ts = ToolSearchToolSet(
        toolsets=[weather_toolset, loose_toolset],
        namespace_descriptions={"weather": "Weather tools"},
    )

    # Initially: tool_search is first (and only)
    tools = await ts.get_tools(mock_run_context)
    assert next(iter(tools.keys())) == "tool_search"

    # After loading namespace: tool_search still first
    await ts.call_tool("tool_search", {"query": "weather"}, mock_run_context, tools["tool_search"])
    tools = await ts.get_tools(mock_run_context)
    assert next(iter(tools.keys())) == "tool_search"
    assert "get_weather" in tools
    assert "get_forecast" in tools

    # After loading loose tools: tool_search still first
    await ts.call_tool("tool_search", {"query": "file"}, mock_run_context, tools["tool_search"])
    tools = await ts.get_tools(mock_run_context)
    assert next(iter(tools.keys())) == "tool_search"


@pytest.mark.anyio
async def test_toolset_namespace_search_loads_all_tools(weather_toolset, mock_run_context):
    """Searching by namespace name should load all tools in that namespace."""
    ts = ToolSearchToolSet(
        toolsets=[weather_toolset],
        namespace_descriptions={"weather": "Weather related tools"},
    )
    tools = await ts.get_tools(mock_run_context)

    result = await ts.call_tool("tool_search", {"query": "weather"}, mock_run_context, tools["tool_search"])
    assert "[weather]" in result
    assert "Found" in result

    # Both tools should now be visible
    tools = await ts.get_tools(mock_run_context)
    assert "get_weather" in tools
    assert "get_forecast" in tools
    assert "tool_search" in tools


@pytest.mark.anyio
async def test_toolset_tool_search_loads_entire_namespace(weather_toolset, mock_run_context):
    """Searching for a specific tool in a namespace loads the entire namespace."""
    ts = ToolSearchToolSet(toolsets=[weather_toolset])
    tools = await ts.get_tools(mock_run_context)

    # Search for forecast specifically
    result = await ts.call_tool("tool_search", {"query": "forecast"}, mock_run_context, tools["tool_search"])
    assert "get_forecast" in result

    # Both tools should be loaded (entire namespace)
    tools = await ts.get_tools(mock_run_context)
    assert "get_weather" in tools
    assert "get_forecast" in tools


@pytest.mark.anyio
async def test_toolset_loose_tools_load_individually(loose_toolset, mock_run_context):
    """Loose tools (no namespace) load individually."""
    ts = ToolSearchToolSet(toolsets=[loose_toolset])
    tools = await ts.get_tools(mock_run_context)

    # Search for view - keyword "file" matches both view and edit descriptions
    # Use "read" which only matches view's description "Read contents of a file"
    result = await ts.call_tool("tool_search", {"query": "read"}, mock_run_context, tools["tool_search"])
    assert "view" in result

    # Only matched tool(s) should be loaded
    tools = await ts.get_tools(mock_run_context)
    assert "view" in tools


@pytest.mark.anyio
async def test_toolset_mixed_namespace_and_loose(weather_toolset, loose_toolset, mock_run_context):
    """Mix of namespace and loose toolsets should work correctly."""
    ts = ToolSearchToolSet(
        toolsets=[weather_toolset, loose_toolset],
        namespace_descriptions={"weather": "Weather tools"},
    )
    tools = await ts.get_tools(mock_run_context)
    assert "tool_search" in tools
    assert len(tools) == 1  # Only tool_search

    # Search for weather -> loads namespace
    await ts.call_tool("tool_search", {"query": "weather"}, mock_run_context, tools["tool_search"])
    tools = await ts.get_tools(mock_run_context)
    assert "get_weather" in tools
    assert "get_forecast" in tools
    assert "view" not in tools
    assert "edit" not in tools

    # Search for file -> loads individual loose tool
    await ts.call_tool("tool_search", {"query": "file"}, mock_run_context, tools["tool_search"])
    tools = await ts.get_tools(mock_run_context)
    assert "view" in tools or "edit" in tools


@pytest.mark.anyio
async def test_toolset_multiple_namespaces(weather_toolset, finance_toolset, mock_run_context):
    """Multiple namespaces should load independently."""
    ts = ToolSearchToolSet(
        toolsets=[weather_toolset, finance_toolset],
        namespace_descriptions={
            "weather": "Weather tools",
            "finance": "Finance tools",
        },
    )
    tools = await ts.get_tools(mock_run_context)

    # Load weather
    await ts.call_tool("tool_search", {"query": "weather"}, mock_run_context, tools["tool_search"])
    tools = await ts.get_tools(mock_run_context)
    assert "get_weather" in tools
    assert "get_forecast" in tools
    assert "get_stock_price" not in tools

    # Load finance
    await ts.call_tool("tool_search", {"query": "finance"}, mock_run_context, tools["tool_search"])
    tools = await ts.get_tools(mock_run_context)
    assert "get_weather" in tools
    assert "get_forecast" in tools
    assert "get_stock_price" in tools
    assert "convert_currency" in tools


@pytest.mark.anyio
async def test_toolset_search_no_results(weather_toolset, mock_run_context):
    """Search with no matches should return helpful message."""
    ts = ToolSearchToolSet(toolsets=[weather_toolset])
    tools = await ts.get_tools(mock_run_context)

    result = await ts.call_tool("tool_search", {"query": "database_migration"}, mock_run_context, tools["tool_search"])
    assert "No tools found" in result


@pytest.mark.anyio
async def test_toolset_search_all_loaded(weather_toolset, mock_run_context):
    """Search when all tools are loaded should say so."""
    # Pre-load the namespace
    mock_run_context.deps.tool_search_loaded_namespaces.append("weather")

    ts = ToolSearchToolSet(toolsets=[weather_toolset])
    tools = await ts.get_tools(mock_run_context)

    result = await ts.call_tool("tool_search", {"query": "anything"}, mock_run_context, tools["tool_search"])
    assert "already loaded" in result


@pytest.mark.anyio
async def test_toolset_call_tool_delegation(weather_toolset, mock_run_context):
    """Calling a discovered tool should delegate to the owning toolset."""
    mock_run_context.deps.tool_search_loaded_namespaces.append("weather")

    ts = ToolSearchToolSet(toolsets=[weather_toolset])
    tools = await ts.get_tools(mock_run_context)

    result = await ts.call_tool("get_weather", {"location": "Tokyo"}, mock_run_context, tools["get_weather"])
    assert "Tokyo" in result
    assert "sunny" in result


@pytest.mark.anyio
async def test_toolset_call_unknown_tool(weather_toolset, mock_run_context):
    """Calling an unknown tool should return an error message."""
    ts = ToolSearchToolSet(toolsets=[weather_toolset])
    tools = await ts.get_tools(mock_run_context)

    result = await ts.call_tool(
        "nonexistent",
        {},
        mock_run_context,
        tools["tool_search"],  # dummy tool
    )
    assert "not found" in result


@pytest.mark.anyio
async def test_toolset_search_empty_query(weather_toolset, mock_run_context):
    """Empty query should return error."""
    ts = ToolSearchToolSet(toolsets=[weather_toolset])
    tools = await ts.get_tools(mock_run_context)

    result = await ts.call_tool("tool_search", {"query": ""}, mock_run_context, tools["tool_search"])
    assert "Error" in result or "required" in result


@pytest.mark.anyio
async def test_toolset_context_manager(weather_toolset, finance_toolset):
    """Context manager should enter/exit wrapped toolsets."""
    ts = ToolSearchToolSet(toolsets=[weather_toolset, finance_toolset])
    async with ts as entered:
        assert entered is ts


@pytest.mark.anyio
async def test_toolset_search_result_includes_params(weather_toolset, mock_run_context):
    """Search results for loose tools should include parameter information."""
    # Use loose toolset for individual tool results
    loose_ts = Toolset(tools=[GetWeatherTool])
    ts = ToolSearchToolSet(toolsets=[loose_ts])
    tools = await ts.get_tools(mock_run_context)

    result = await ts.call_tool("tool_search", {"query": "weather"}, mock_run_context, tools["tool_search"])
    assert "location" in result
    assert "Parameters" in result


# ---------------------------------------------------------------------------
# State persistence tests (session restore via AgentContext)
# ---------------------------------------------------------------------------


@pytest.mark.anyio
async def test_state_persists_in_context(weather_toolset, mock_run_context):
    """Loaded state should be stored in AgentContext fields."""
    ts = ToolSearchToolSet(toolsets=[weather_toolset])
    tools = await ts.get_tools(mock_run_context)

    assert mock_run_context.deps.tool_search_loaded_namespaces == []
    assert mock_run_context.deps.tool_search_loaded_tools == []

    await ts.call_tool("tool_search", {"query": "weather"}, mock_run_context, tools["tool_search"])

    assert "weather" in mock_run_context.deps.tool_search_loaded_namespaces


@pytest.mark.anyio
async def test_state_restore_from_context(weather_toolset, finance_toolset, mock_run_context):
    """Pre-populated context should restore loaded tools."""
    mock_run_context.deps.tool_search_loaded_namespaces = ["weather"]

    ts = ToolSearchToolSet(toolsets=[weather_toolset, finance_toolset])
    tools = await ts.get_tools(mock_run_context)

    # Weather should be visible, finance should not
    assert "get_weather" in tools
    assert "get_forecast" in tools
    assert "get_stock_price" not in tools
    assert "convert_currency" not in tools


@pytest.mark.anyio
async def test_state_restore_loose_tools(loose_toolset, mock_run_context):
    """Pre-populated loose tools should be restored."""
    mock_run_context.deps.tool_search_loaded_tools = ["view"]

    ts = ToolSearchToolSet(toolsets=[loose_toolset])
    tools = await ts.get_tools(mock_run_context)

    assert "view" in tools
    assert "edit" not in tools


@pytest.mark.anyio
async def test_export_and_restore_state(weather_toolset, finance_toolset):
    """Full round-trip: export state -> restore -> tools visible."""
    ctx1 = AgentContext(run_id="session-1")
    run_ctx1 = MagicMock(spec=RunContext)
    run_ctx1.deps = ctx1

    ts = ToolSearchToolSet(
        toolsets=[weather_toolset, finance_toolset],
        namespace_descriptions={"weather": "Weather tools", "finance": "Finance tools"},
    )

    # Load weather namespace
    tools = await ts.get_tools(run_ctx1)
    await ts.call_tool("tool_search", {"query": "weather"}, run_ctx1, tools["tool_search"])

    # Export state
    state = ctx1.export_state()
    assert "weather" in state.tool_search_loaded_namespaces

    # Restore in new context
    ctx2 = AgentContext(run_id="session-2").with_state(state)
    run_ctx2 = MagicMock(spec=RunContext)
    run_ctx2.deps = ctx2

    tools = await ts.get_tools(run_ctx2)
    assert "get_weather" in tools
    assert "get_forecast" in tools
    assert "get_stock_price" not in tools


# ---------------------------------------------------------------------------
# Namespace description resolution tests
# ---------------------------------------------------------------------------


@pytest.mark.anyio
async def test_namespace_description_from_explicit(weather_toolset, mock_run_context):
    """namespace_descriptions takes priority over toolset.description."""
    ts = ToolSearchToolSet(
        toolsets=[weather_toolset],
        namespace_descriptions={"weather": "Custom weather description"},
    )
    await ts.get_tools(mock_run_context)
    instruction = await ts.get_instructions(mock_run_context)
    assert instruction is not None
    assert "Custom weather description" in instruction


@pytest.mark.anyio
async def test_namespace_description_from_toolset_description(mock_run_context):
    """Toolset.description should be used when namespace_descriptions is not provided."""
    ts_with_desc = Toolset(
        tools=[GetWeatherTool],
        toolset_id="weather",
        description="Weather toolset from description property",
    )
    ts = ToolSearchToolSet(toolsets=[ts_with_desc])
    await ts.get_tools(mock_run_context)
    instruction = await ts.get_instructions(mock_run_context)
    assert instruction is not None
    assert "Weather toolset from description property" in instruction


@pytest.mark.anyio
async def test_namespace_description_auto_fallback(mock_run_context):
    """Without any description source, auto-generate from toolset id."""
    ts_no_desc = Toolset(tools=[GetWeatherTool], toolset_id="weather")
    ts = ToolSearchToolSet(toolsets=[ts_no_desc])
    await ts.get_tools(mock_run_context)
    instruction = await ts.get_instructions(mock_run_context)
    assert instruction is not None
    assert "weather" in instruction


@pytest.mark.anyio
async def test_namespace_description_from_instructions_property(mock_run_context):
    """Toolsets with an 'instructions' property (e.g., MCPServer) use first line as description."""
    ts_with_instructions = Toolset(tools=[GetWeatherTool], toolset_id="mcp_server")
    # Simulate MCPServer.instructions property (available after __aenter__)
    ts_with_instructions.instructions = "MCP server for weather data\nMore details here"  # type: ignore[attr-defined]

    ts = ToolSearchToolSet(toolsets=[ts_with_instructions])
    await ts.get_tools(mock_run_context)
    instruction = await ts.get_instructions(mock_run_context)
    assert instruction is not None
    assert "MCP server for weather data" in instruction
    # Should not include the second line
    assert "More details here" not in instruction


@pytest.mark.anyio
async def test_namespace_description_instructions_lower_priority_than_explicit(mock_run_context):
    """namespace_descriptions should take priority over instructions property."""
    ts_with_instructions = Toolset(tools=[GetWeatherTool], toolset_id="mcp_server")
    ts_with_instructions.instructions = "From MCP instructions"  # type: ignore[attr-defined]

    ts = ToolSearchToolSet(
        toolsets=[ts_with_instructions],
        namespace_descriptions={"mcp_server": "User-provided description"},
    )
    await ts.get_tools(mock_run_context)
    instruction = await ts.get_instructions(mock_run_context)
    assert instruction is not None
    assert "User-provided description" in instruction
    assert "From MCP instructions" not in instruction


# ---------------------------------------------------------------------------
# Instructions tests
# ---------------------------------------------------------------------------


@pytest.mark.anyio
async def test_instructions_include_namespace_info(weather_toolset, finance_toolset, mock_run_context):
    """Instructions should list available namespaces."""
    ts = ToolSearchToolSet(
        toolsets=[weather_toolset, finance_toolset],
        namespace_descriptions={
            "weather": "Weather tools",
            "finance": "Finance tools",
        },
    )
    await ts.get_tools(mock_run_context)
    instruction = await ts.get_instructions(mock_run_context)
    assert instruction is not None
    assert "weather" in instruction
    assert "finance" in instruction
    assert "namespaces" in instruction.lower()


@pytest.mark.anyio
async def test_instructions_only_for_loaded_toolsets(weather_toolset, mock_run_context):
    """Wrapped toolset instructions should only be included when loaded."""
    ts = ToolSearchToolSet(toolsets=[weather_toolset])

    # No tools loaded yet - should not include weather toolset instructions
    await ts.get_tools(mock_run_context)
    instruction = await ts.get_instructions(mock_run_context)
    # Should only have tool_search instruction, not weather toolset instructions
    assert instruction is not None
    assert "tool_search" in instruction


# ---------------------------------------------------------------------------
# EmbeddingSearchStrategy tests (requires fastembed + numpy)
# ---------------------------------------------------------------------------


fastembed = pytest.importorskip("fastembed", reason="fastembed not installed")
np = pytest.importorskip("numpy", reason="numpy not installed")


@pytest.fixture(scope="module")
def embedding_strategy():
    """Create an EmbeddingSearchStrategy (slow model init, reuse across tests).

    Skips if model download or loading fails (e.g., in CI without network
    access or with incomplete model cache).
    """
    from ya_agent_sdk.toolsets.tool_search.strategies.embedding import EmbeddingSearchStrategy

    strategy = EmbeddingSearchStrategy()
    try:
        # Trigger model download/load and a test embed to detect failures early
        model = strategy._get_model()
        list(model.embed(["test"]))
    except Exception as exc:
        pytest.skip(f"Embedding model not available: {exc}")
    return strategy


@pytest.fixture
async def indexed_candidates(embedding_strategy):
    """Build an embedding index with test tools."""
    candidates = [
        ToolMetadata(
            name="get_weather",
            description="Get the current weather in a given location",
            parameter_names=["location", "unit"],
            parameter_descriptions={"location": "The city and state", "unit": "Temperature unit"},
        ),
        ToolMetadata(
            name="get_forecast",
            description="Get the weather forecast for multiple days ahead",
            parameter_names=["location", "days"],
            parameter_descriptions={"location": "The city name", "days": "Number of days to forecast"},
        ),
        ToolMetadata(
            name="get_stock_price",
            description="Get the current stock price for a ticker symbol",
            parameter_names=["ticker"],
            parameter_descriptions={"ticker": "Stock ticker symbol like AAPL"},
        ),
        ToolMetadata(
            name="convert_currency",
            description="Convert an amount from one currency to another using exchange rates",
            parameter_names=["amount", "from_currency", "to_currency"],
            parameter_descriptions={
                "amount": "Amount to convert",
                "from_currency": "Source currency code",
                "to_currency": "Target currency code",
            },
        ),
        ToolMetadata(
            name="send_email",
            description="Send an email message to a recipient",
            parameter_names=["recipient", "subject", "body"],
            parameter_descriptions={
                "recipient": "Email address",
                "subject": "Email subject line",
                "body": "Email body content",
            },
        ),
    ]
    await embedding_strategy.build_index(candidates)
    return candidates


@pytest.mark.anyio
async def test_embedding_search_weather(embedding_strategy, indexed_candidates):
    """Semantic search for 'weather' should rank weather tools first."""
    results = await embedding_strategy.search("check the weather", indexed_candidates, max_results=3)
    assert len(results) >= 1
    result_names = [r.name for r in results]
    assert "get_weather" in result_names or "get_forecast" in result_names


@pytest.mark.anyio
async def test_embedding_search_finance(embedding_strategy, indexed_candidates):
    """Semantic search for finance should rank finance tools first."""
    results = await embedding_strategy.search("stock market price", indexed_candidates, max_results=3)
    assert len(results) >= 1
    assert results[0].name == "get_stock_price"


@pytest.mark.anyio
async def test_embedding_search_currency(embedding_strategy, indexed_candidates):
    """Semantic search for currency conversion."""
    results = await embedding_strategy.search("convert dollars to euros", indexed_candidates, max_results=3)
    assert len(results) >= 1
    assert results[0].name == "convert_currency"


@pytest.mark.anyio
async def test_embedding_search_email(embedding_strategy, indexed_candidates):
    """Semantic search for email should find send_email."""
    results = await embedding_strategy.search("send a message to someone", indexed_candidates, max_results=3)
    assert len(results) >= 1
    assert results[0].name == "send_email"


@pytest.mark.anyio
async def test_embedding_search_max_results(embedding_strategy, indexed_candidates):
    """Should respect max_results parameter."""
    results = await embedding_strategy.search("tool", indexed_candidates, max_results=2)
    assert len(results) <= 2


@pytest.mark.anyio
async def test_embedding_search_empty_query(embedding_strategy, indexed_candidates):
    """Empty query should return empty results."""
    assert await embedding_strategy.search("", indexed_candidates) == []


@pytest.mark.anyio
async def test_embedding_search_empty_candidates(embedding_strategy):
    """Empty candidates should return empty results."""
    assert await embedding_strategy.search("weather", []) == []


@pytest.mark.anyio
async def test_embedding_search_candidate_filtering(embedding_strategy, indexed_candidates):
    """Should only return tools from the candidates list, not all indexed tools."""
    weather_only = [t for t in indexed_candidates if "weather" in t.name or "forecast" in t.name]
    results = await embedding_strategy.search("stock price", weather_only, max_results=5)
    result_names = {r.name for r in results}
    assert "get_stock_price" not in result_names


@pytest.mark.anyio
async def test_embedding_build_index_empty(embedding_strategy):
    """Building index with empty list should not crash."""
    await embedding_strategy.build_index([])
    results = await embedding_strategy.search("anything", [])
    assert results == []


@pytest.mark.anyio
async def test_embedding_build_index_rebuild(embedding_strategy, indexed_candidates):
    """Rebuilding index should replace previous index."""
    new_tools = [
        ToolMetadata(name="ping", description="Ping a server to check if it is alive"),
    ]
    await embedding_strategy.build_index(new_tools)
    results = await embedding_strategy.search("ping server", new_tools, max_results=3)
    assert len(results) == 1
    assert results[0].name == "ping"

    # Rebuild with original for other tests
    await embedding_strategy.build_index(indexed_candidates)


@pytest.mark.anyio
async def test_toolset_with_embedding_strategy(weather_toolset, finance_toolset, mock_run_context, embedding_strategy):
    """ToolSearchToolSet should work with EmbeddingSearchStrategy."""
    from ya_agent_sdk.toolsets.tool_search.strategies.embedding import EmbeddingSearchStrategy

    ts = ToolSearchToolSet(
        toolsets=[weather_toolset, finance_toolset],
        namespace_descriptions={"weather": "Weather tools", "finance": "Finance tools"},
        search_strategy=EmbeddingSearchStrategy(),
    )
    tools = await ts.get_tools(mock_run_context)
    assert "tool_search" in tools
    assert "get_weather" not in tools

    # Search for finance tools
    result = await ts.call_tool("tool_search", {"query": "stock market"}, mock_run_context, tools["tool_search"])
    assert "get_stock_price" in result or "finance" in result

    # Finance namespace should be loaded
    tools = await ts.get_tools(mock_run_context)
    assert "get_stock_price" in tools
    assert "convert_currency" in tools


# ---------------------------------------------------------------------------
# optional_namespaces tests
# ---------------------------------------------------------------------------


class FailingToolset(Toolset):
    """A toolset that raises on __aenter__ to simulate connection failure."""

    async def __aenter__(self):
        raise ConnectionError("Failed to connect to MCP server")


@pytest.mark.anyio
async def test_optional_namespace_skipped_on_failure(mock_run_context):
    """Optional toolset that fails to init should be skipped."""
    good_toolset = Toolset(tools=[GetWeatherTool, GetForecastTool], toolset_id="weather")
    bad_toolset = FailingToolset(tools=[GetStockPriceTool], toolset_id="failing_mcp")

    ts = ToolSearchToolSet(
        toolsets=[good_toolset, bad_toolset],
        optional_namespaces={"failing_mcp"},
    )

    # Should not raise, the failing toolset is optional
    async with ts:
        tools = await ts.get_tools(mock_run_context)
        assert "tool_search" in tools

        # Load weather namespace - should work
        await ts.call_tool("tool_search", {"query": "weather"}, mock_run_context, tools["tool_search"])
        tools = await ts.get_tools(mock_run_context)
        assert "get_weather" in tools
        assert "get_forecast" in tools


@pytest.mark.anyio
async def test_required_namespace_raises_on_failure(mock_run_context):
    """Required toolset (not in optional_namespaces) should raise on init failure."""
    good_toolset = Toolset(tools=[GetWeatherTool], toolset_id="weather")
    bad_toolset = FailingToolset(tools=[GetStockPriceTool], toolset_id="critical_mcp")

    ts = ToolSearchToolSet(
        toolsets=[good_toolset, bad_toolset],
        optional_namespaces=set(),  # No optional namespaces
    )

    with pytest.raises(ConnectionError, match="Failed to connect"):
        async with ts:
            pass


@pytest.mark.anyio
async def test_required_namespace_default_raises_on_failure(mock_run_context):
    """Without optional_namespaces parameter, all toolsets are required (backward compat)."""
    bad_toolset = FailingToolset(tools=[GetStockPriceTool], toolset_id="some_mcp")

    ts = ToolSearchToolSet(toolsets=[bad_toolset])

    with pytest.raises(ConnectionError, match="Failed to connect"):
        async with ts:
            pass


@pytest.mark.anyio
async def test_optional_namespace_no_id_still_raises():
    """Toolset without id that fails should always raise (can't match optional_namespaces)."""
    bad_toolset = FailingToolset(tools=[GetStockPriceTool])  # No toolset_id

    ts = ToolSearchToolSet(
        toolsets=[bad_toolset],
        optional_namespaces=set(),
    )

    with pytest.raises(ConnectionError, match="Failed to connect"):
        async with ts:
            pass


@pytest.mark.anyio
async def test_optional_namespace_rollback_on_required_failure(mock_run_context):
    """When a required toolset fails, already-entered optional toolsets are rolled back."""
    good_toolset = Toolset(tools=[GetWeatherTool], toolset_id="weather")
    bad_toolset = FailingToolset(tools=[GetStockPriceTool], toolset_id="critical_mcp")

    ts = ToolSearchToolSet(
        toolsets=[good_toolset, bad_toolset],
        optional_namespaces=set(),  # Both required
    )

    with pytest.raises(ConnectionError):
        async with ts:
            pass


@pytest.mark.anyio
async def test_multiple_optional_namespaces_some_fail(mock_run_context):
    """Multiple optional toolsets: some fail, some succeed."""
    good_weather = Toolset(tools=[GetWeatherTool, GetForecastTool], toolset_id="weather")
    bad_mcp1 = FailingToolset(tools=[GetStockPriceTool], toolset_id="mcp1")
    bad_mcp2 = FailingToolset(tools=[ConvertCurrencyTool], toolset_id="mcp2")

    ts = ToolSearchToolSet(
        toolsets=[good_weather, bad_mcp1, bad_mcp2],
        optional_namespaces={"mcp1", "mcp2"},
    )

    async with ts:
        tools = await ts.get_tools(mock_run_context)
        assert "tool_search" in tools

        # Weather should still work
        await ts.call_tool("tool_search", {"query": "weather"}, mock_run_context, tools["tool_search"])
        tools = await ts.get_tools(mock_run_context)
        assert "get_weather" in tools


# ---------------------------------------------------------------------------
# init_report and ToolSearchInitEvent tests
# ---------------------------------------------------------------------------


@pytest.mark.anyio
async def test_init_report_all_connected(mock_run_context):
    """init_report shows connected status for all successfully initialized namespaces."""
    from ya_agent_sdk.events import NamespaceStatus

    weather = Toolset(tools=[GetWeatherTool], toolset_id="weather")
    finance = Toolset(tools=[GetStockPriceTool], toolset_id="finance")

    ts = ToolSearchToolSet(toolsets=[weather, finance])

    async with ts:
        report = ts.init_report
        assert report == {
            "weather": NamespaceStatus.connected,
            "finance": NamespaceStatus.connected,
        }


@pytest.mark.anyio
async def test_init_report_mixed_status(mock_run_context):
    """init_report shows skipped status for failed optional namespaces."""
    from ya_agent_sdk.events import NamespaceStatus

    good = Toolset(tools=[GetWeatherTool], toolset_id="weather")
    bad = FailingToolset(tools=[GetStockPriceTool], toolset_id="bad_mcp")

    ts = ToolSearchToolSet(
        toolsets=[good, bad],
        optional_namespaces={"bad_mcp"},
    )

    async with ts:
        report = ts.init_report
        assert report == {
            "weather": NamespaceStatus.connected,
            "bad_mcp": NamespaceStatus.skipped,
        }


@pytest.mark.anyio
async def test_init_report_empty_without_namespaces():
    """Toolsets without id are not tracked in init_report."""
    loose = Toolset(tools=[GetWeatherTool])  # No toolset_id

    ts = ToolSearchToolSet(toolsets=[loose])
    async with ts:
        assert ts.init_report == {}


@pytest.mark.anyio
async def test_init_report_is_copy():
    """init_report returns a copy, not a reference to internal state."""
    from ya_agent_sdk.events import NamespaceStatus

    weather = Toolset(tools=[GetWeatherTool], toolset_id="weather")
    ts = ToolSearchToolSet(toolsets=[weather])

    async with ts:
        report1 = ts.init_report
        report1["weather"] = NamespaceStatus.skipped  # Mutate the copy
        assert ts.init_report["weather"] == NamespaceStatus.connected  # Original unchanged


@pytest.mark.anyio
async def test_init_event_emitted_on_first_get_tools(mock_run_context):
    """ToolSearchInitEvent is emitted on get_tools call."""
    weather = Toolset(tools=[GetWeatherTool], toolset_id="weather")
    ts = ToolSearchToolSet(toolsets=[weather])

    async with ts:
        # Enable streaming to capture the event
        mock_run_context.deps._stream_queue_enabled = True
        await ts.get_tools(mock_run_context)

        # Check event was placed on the queue
        queue = mock_run_context.deps.agent_stream_queues[mock_run_context.deps._agent_id]
        assert not queue.empty()
        event = queue.get_nowait()

        from ya_agent_sdk.events import NamespaceStatus, ToolSearchInitEvent

        assert isinstance(event, ToolSearchInitEvent)
        assert event.namespace_status == {"weather": NamespaceStatus.connected}


@pytest.mark.anyio
async def test_init_event_emitted_every_call(mock_run_context):
    """ToolSearchInitEvent is emitted on every get_tools call (status can change)."""
    weather = Toolset(tools=[GetWeatherTool], toolset_id="weather")
    ts = ToolSearchToolSet(toolsets=[weather])

    async with ts:
        mock_run_context.deps._stream_queue_enabled = True
        await ts.get_tools(mock_run_context)
        await ts.get_tools(mock_run_context)

        # Two events should be on the queue
        queue = mock_run_context.deps.agent_stream_queues[mock_run_context.deps._agent_id]
        assert queue.qsize() == 2


@pytest.mark.anyio
async def test_init_event_not_emitted_without_namespaces(mock_run_context):
    """No event emitted if there are no namespace toolsets."""
    loose = Toolset(tools=[GetWeatherTool])  # No toolset_id
    ts = ToolSearchToolSet(toolsets=[loose])

    async with ts:
        mock_run_context.deps._stream_queue_enabled = True
        await ts.get_tools(mock_run_context)

        # No event because init_report is empty
        queue = mock_run_context.deps.agent_stream_queues[mock_run_context.deps._agent_id]
        assert queue.empty()


class RuntimeFailingToolset(Toolset):
    """A toolset that initializes OK but fails on get_tools (simulates runtime disconnect)."""

    _fail_get_tools: bool = False

    async def get_tools(self, ctx):
        if self._fail_get_tools:
            raise ConnectionError("MCP server disconnected")
        return await super().get_tools(ctx)


@pytest.mark.anyio
async def test_optional_namespace_runtime_error(mock_run_context):
    """Optional namespace that fails during get_tools should be skipped with error status."""
    from ya_agent_sdk.events import NamespaceStatus

    good = Toolset(tools=[GetWeatherTool], toolset_id="weather")
    flaky = RuntimeFailingToolset(tools=[GetStockPriceTool], toolset_id="flaky_mcp")

    ts = ToolSearchToolSet(
        toolsets=[good, flaky],
        optional_namespaces={"flaky_mcp"},
    )

    async with ts:
        # First call: both work
        tools = await ts.get_tools(mock_run_context)
        assert ts.init_report["flaky_mcp"] == NamespaceStatus.connected

        # Simulate runtime disconnect
        flaky._fail_get_tools = True
        tools = await ts.get_tools(mock_run_context)

        # flaky_mcp should now be error, weather still connected
        assert ts.init_report == {
            "weather": NamespaceStatus.connected,
            "flaky_mcp": NamespaceStatus.error,
        }
        # tool_search and weather tools should still work
        assert "tool_search" in tools


@pytest.mark.anyio
async def test_required_namespace_runtime_error_raises(mock_run_context):
    """Required namespace that fails during get_tools should raise."""
    flaky = RuntimeFailingToolset(tools=[GetStockPriceTool], toolset_id="critical_mcp")
    flaky._fail_get_tools = True

    ts = ToolSearchToolSet(toolsets=[flaky])

    async with ts:
        with pytest.raises(ConnectionError, match="MCP server disconnected"):
            await ts.get_tools(mock_run_context)



"""Tests for document conversion tools (PDF and Office)."""

import shutil
from contextlib import AsyncExitStack
from pathlib import Path
from unittest.mock import MagicMock

import pytest
from inline_snapshot import snapshot
from pydantic_ai import RunContext

from ya_agent_sdk.context import AgentContext
from ya_agent_sdk.environment.local import LocalEnvironment

# Test file directory
TEST_FILES_DIR = Path(__file__).parent


@pytest.fixture
def pdf_file(tmp_path: Path) -> Path:
    """Copy dummy.pdf to tmp_path for testing."""
    src = TEST_FILES_DIR / "dummy.pdf"
    if src.exists():
        dst = tmp_path / "dummy.pdf"
        shutil.copy(src, dst)
        return dst
    pytest.skip("dummy.pdf not found")


@pytest.fixture
def docx_file(tmp_path: Path) -> Path:
    """Copy dummy.docx to tmp_path for testing."""
    src = TEST_FILES_DIR / "dummy.docx"
    if src.exists():
        dst = tmp_path / "dummy.docx"
        shutil.copy(src, dst)
        return dst
    pytest.skip("dummy.docx not found")


# --- PDF Convert Tool Tests ---

# Check if PDF tools are available
try:
    from ya_agent_sdk.toolsets.core.document.pdf import PdfConvertTool

    PDF_AVAILABLE = True
except ImportError:
    PDF_AVAILABLE = False


@pytest.mark.skipif(not PDF_AVAILABLE, reason="pymupdf not installed")
def test_pdf_convert_tool_attributes(agent_context: AgentContext) -> None:
    """Should have correct name and description."""
    tool = PdfConvertTool()
    assert tool.name == "pdf_convert"
    assert "PDF" in tool.description or "pdf" in tool.description.lower()


@pytest.mark.skipif(not PDF_AVAILABLE, reason="pymupdf not installed")
async def test_pdf_convert_file_not_found(tmp_path: Path) -> None:
    """Should return error when file not found."""
    async with AsyncExitStack() as stack:
        env = await stack.enter_async_context(
            LocalEnvironment(allowed_paths=[tmp_path], default_path=tmp_path, tmp_base_dir=tmp_path)
        )
        ctx = await stack.enter_async_context(AgentContext(env=env))
        tool = PdfConvertTool()

        mock_run_ctx = MagicMock(spec=RunContext)
        mock_run_ctx.deps = ctx

        result = await tool.call(mock_run_ctx, file_path="nonexistent.pdf")
        assert result["success"] is False
        assert "not found" in result["error"].lower()


@pytest.mark.skipif(not PDF_AVAILABLE, reason="pymupdf not installed")
async def test_pdf_convert_not_pdf_file(tmp_path: Path) -> None:
    """Should return error when file is not PDF."""
    async with AsyncExitStack() as stack:
        env = await stack.enter_async_context(
            LocalEnvironment(allowed_paths=[tmp_path], default_path=tmp_path, tmp_base_dir=tmp_path)
        )
        ctx = await stack.enter_async_context(AgentContext(env=env))
        tool = PdfConvertTool()

        # Create a non-PDF file
        txt_file = tmp_path / "test.txt"
        txt_file.write_text("not a pdf")

        mock_run_ctx = MagicMock(spec=RunContext)
        mock_run_ctx.deps = ctx

        result = await tool.call(mock_run_ctx, file_path="test.txt")
        assert result["success"] is False
        assert "not a pdf" in result["error"].lower()


@pytest.mark.skipif(not PDF_AVAILABLE, reason="pymupdf not installed")
async def test_pdf_convert_success(tmp_path: Path, pdf_file: Path) -> None:
    """Should successfully convert PDF to markdown."""
    async with AsyncExitStack() as stack:
        env = await stack.enter_async_context(
            LocalEnvironment(allowed_paths=[tmp_path], default_path=tmp_path, tmp_base_dir=tmp_path)
        )
        ctx = await stack.enter_async_context(AgentContext(env=env))
        tool = PdfConvertTool()

        mock_run_ctx = MagicMock(spec=RunContext)
        mock_run_ctx.deps = ctx

        result = await tool.call(mock_run_ctx, file_path="dummy.pdf")
        assert result["success"] is True
        assert result == snapshot({
            "success": True,
            "export_path": "export_dummy",
            "markdown_path": "export_dummy/dummy.md",
            "total_pages": 1,
            "converted_pages": 1,
            "page_range": "1-1",
        })

        # Verify export directory was created
        export_dir = tmp_path / result["export_path"]
        assert export_dir.exists()
        assert (tmp_path / result["markdown_path"]).exists()


@pytest.mark.skipif(not PDF_AVAILABLE, reason="pymupdf not installed")
async def test_pdf_convert_page_range(tmp_path: Path, pdf_file: Path) -> None:
    """Should respect page range parameters."""
    async with AsyncExitStack() as stack:
        env = await stack.enter_async_context(
            LocalEnvironment(allowed_paths=[tmp_path], default_path=tmp_path, tmp_base_dir=tmp_path)
        )
        ctx = await stack.enter_async_context(AgentContext(env=env))
        tool = PdfConvertTool()

        mock_run_ctx = MagicMock(spec=RunContext)
        mock_run_ctx.deps = ctx

        result = await tool.call(mock_run_ctx, file_path="dummy.pdf", page_start=1, page_end=1)
        assert result["success"] is True
        assert result["converted_pages"] == 1
        assert result == snapshot({
            "success": True,
            "export_path": "export_dummy",
            "markdown_path": "export_dummy/dummy.md",
            "total_pages": 1,
            "converted_pages": 1,
            "page_range": "1-1",
        })

        # Verify export directory was created
        export_dir = tmp_path / result["export_path"]
        assert export_dir.exists()
        assert (tmp_path / result["markdown_path"]).exists()
        assert (tmp_path / result["markdown_path"]).read_text() == snapshot("""\
**Dummy PDF file** \n\

""")


# --- Office Convert Tool Tests ---

# Check if Office tools are available
try:
    from ya_agent_sdk.toolsets.core.document.office import OfficeConvertTool

    OFFICE_AVAILABLE = True
except ImportError:
    OFFICE_AVAILABLE = False


@pytest.mark.skipif(not OFFICE_AVAILABLE, reason="markitdown not installed")
def test_office_convert_tool_attributes(agent_context: AgentContext) -> None:
    """Should have correct name and description."""
    tool = OfficeConvertTool()
    assert tool.name == "office_to_markdown"
    assert "Office" in tool.description or "Word" in tool.description


@pytest.mark.skipif(not OFFICE_AVAILABLE, reason="markitdown not installed")
async def test_office_convert_file_not_found(tmp_path: Path) -> None:
    """Should return error when file not found."""
    async with AsyncExitStack() as stack:
        env = await stack.enter_async_context(
            LocalEnvironment(allowed_paths=[tmp_path], default_path=tmp_path, tmp_base_dir=tmp_path)
        )
        ctx = await stack.enter_async_context(AgentContext(env=env))
        tool = OfficeConvertTool()

        mock_run_ctx = MagicMock(spec=RunContext)
        mock_run_ctx.deps = ctx

        result = await tool.call(mock_run_ctx, file_path="nonexistent.docx")
        assert result["success"] is False
        assert "not found" in result["error"].lower()


@pytest.mark.skipif(not OFFICE_AVAILABLE, reason="markitdown not installed")
async def test_office_convert_unsupported_format(tmp_path: Path) -> None:
    """Should return error for unsupported formats."""
    async with AsyncExitStack() as stack:
        env = await stack.enter_async_context(
            LocalEnvironment(allowed_paths=[tmp_path], default_path=tmp_path, tmp_base_dir=tmp_path)
        )
        ctx = await stack.enter_async_context(AgentContext(env=env))
        tool = OfficeConvertTool()

        # Create an unsupported file
        txt_file = tmp_path / "test.txt"
        txt_file.write_text("not an office doc")

        mock_run_ctx = MagicMock(spec=RunContext)
        mock_run_ctx.deps = ctx

        result = await tool.call(mock_run_ctx, file_path="test.txt")
        assert result["success"] is False
        assert "unsupported" in result["error"].lower()


@pytest.mark.skipif(not OFFICE_AVAILABLE, reason="markitdown not installed")
async def test_office_convert_docx_success(tmp_path: Path, docx_file: Path) -> None:
    """Should successfully convert DOCX to markdown."""
    async with AsyncExitStack() as stack:
        env = await stack.enter_async_context(
            LocalEnvironment(allowed_paths=[tmp_path], default_path=tmp_path, tmp_base_dir=tmp_path)
        )
        ctx = await stack.enter_async_context(AgentContext(env=env))
        tool = OfficeConvertTool()

        mock_run_ctx = MagicMock(spec=RunContext)
        mock_run_ctx.deps = ctx

        result = await tool.call(mock_run_ctx, file_path="dummy.docx")
        assert result["success"] is True
        assert result == snapshot({
            "success": True,
            "export_path": "export_dummy",
            "markdown_path": "export_dummy/dummy.md",
        })

        # Verify export directory was created
        export_dir = tmp_path / result["export_path"]
        assert export_dir.exists()
        assert (tmp_path / result["markdown_path"]).exists()
        assert "这是一首简单的小情歌" in (tmp_path / result["markdown_path"]).read_text()
        assert "images/" in (tmp_path / result["markdown_path"]).read_text()



"""List tool for directory listing."""

import fnmatch
from functools import cache
from pathlib import Path
from typing import Annotated, Any, cast

from pydantic import Field
from pydantic_ai import RunContext
from y_agent_environment import FileOperator

from ya_agent_sdk._logger import get_logger
from ya_agent_sdk.context import AgentContext
from ya_agent_sdk.toolsets.core.base import BaseTool
from ya_agent_sdk.toolsets.core.filesystem._types import FileInfoWithStats

logger = get_logger(__name__)

_PROMPTS_DIR = Path(__file__).parent / "prompts"


@cache
def _load_instruction() -> str:
    """Load ls instruction from prompts/ls.md."""
    prompt_file = _PROMPTS_DIR / "ls.md"
    return prompt_file.read_text()


class ListTool(BaseTool):
    """Tool for listing files and directories."""

    name = "ls"
    description = "List directory contents with file info (name, type, size, modified time)."

    def is_available(self, ctx: RunContext[AgentContext]) -> bool:
        """Check if tool is available (requires file_operator)."""
        if ctx.deps.file_operator is None:
            logger.debug("ListTool unavailable: file_operator is not configured")
            return False
        return True

    async def get_instruction(self, ctx: RunContext[AgentContext]) -> str | None:
        """Load instruction from prompts/ls.md."""
        return _load_instruction()

    async def call(
        self,
        ctx: RunContext[AgentContext],
        path: Annotated[
            str,
            Field(description="Directory relative path"),
        ],
        ignore: Annotated[
            list[str] | None,
            Field(default=None, description="Glob patterns to ignore"),
        ] = None,
    ) -> dict[str, Any]:
        """List directory contents with detailed information."""
        file_operator = cast(FileOperator, ctx.deps.file_operator)

        if not await file_operator.exists(path):
            return {"success": False, "error": f"Directory not found: {path}"}

        if not await file_operator.is_dir(path):
            return {"success": False, "error": f"Path is not a directory: {path}"}

        def should_ignore(name: str) -> bool:
            if not ignore:
                return False
            return any(fnmatch.fnmatch(name, pattern) for pattern in ignore)

        entries: list[FileInfoWithStats] = []

        try:
            items = await file_operator.list_dir(path)

            for name in items:
                if should_ignore(name):
                    continue

                item_path = f"{path}/{name}" if path != "." else name
                is_dir = await file_operator.is_dir(item_path)

                file_info: FileInfoWithStats = {
                    "name": name,
                    "path": item_path,
                    "type": "directory" if is_dir else "file",
                }

                if not is_dir:
                    try:
                        stat = await file_operator.stat(item_path)
                        file_info["size"] = stat["size"]
                        file_info["modified"] = stat["mtime"]
                    except Exception as e:
                        file_info["error"] = f"Failed to get file stats: {e!s}"

                entries.append(file_info)

        except Exception as e:
            return {"success": False, "error": f"Failed to list directory: {e!s}"}

        return {
            "path": path,
            "entries": entries,
            "count": len(entries),
            "success": True,
        }


__all__ = ["ListTool"]
